# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines functions and classes for building and manipulating types."""

import abc
import atexit
import collections
from collections.abc import Hashable, Iterable, Iterator, Mapping, MutableMapping, Sequence
from typing import Optional, TypeVar, Union
import weakref

import attrs
from federated_language.common_libs import py_typecheck
from federated_language.common_libs import structure
from federated_language.proto import array_pb2
from federated_language.proto import computation_pb2
from federated_language.types import array_shape
from federated_language.types import dtype_utils
from federated_language.types import placements
import numpy as np
from typing_extensions import TypeGuard


T = TypeVar('T')


class UnexpectedTypeError(TypeError):

  def __init__(self, expected: type['Type'], actual: 'Type'):
    super().__init__(f'Expected type of kind {expected}, found type {actual}.')


class TypeNotAssignableError(TypeError):

  def __init__(self, source_type, target_type):
    super().__init__(
        f'Expected {source_type} to be assignable to {target_type}.'
    )


class TypesNotEquivalentError(TypeError):

  def __init__(self, first_type, second_type):
    super().__init__(
        f'Expected {first_type} to be equivalent to {second_type}.'
    )


def _check_type_has_field(type_pb: computation_pb2.Type, field: str):
  if not type_pb.HasField(field):
    raise ValueError(
        f'Expected `type_pb` to have the field "{field}", found {type_pb}.'
    )


class Type(abc.ABC):
  """An abstract interface for all classes that represent types."""

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'Type':
    """Returns a `Type` for the `type_pb`."""
    type_oneof = type_pb.WhichOneof('type')
    if type_oneof == 'federated':
      return FederatedType.from_proto(type_pb)
    elif type_oneof == 'function':
      return FunctionType.from_proto(type_pb)
    elif type_oneof == 'placement':
      return PlacementType.from_proto(type_pb)
    elif type_oneof == 'sequence':
      return SequenceType.from_proto(type_pb)
    elif type_oneof == 'struct':
      return StructType.from_proto(type_pb)
    elif type_oneof == 'tensor':
      return TensorType.from_proto(type_pb)
    else:
      raise NotImplementedError(f'Unexpected type found: {type_oneof}.')

  @abc.abstractmethod
  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    raise NotImplementedError

  def compact_representation(self) -> str:
    """Returns the compact string representation of this type."""
    return _string_representation(self, formatted=False)

  def formatted_representation(self) -> str:
    """Returns the formatted string representation of this type."""
    return _string_representation(self, formatted=True)

  @abc.abstractmethod
  def children(self) -> Iterator['Type']:
    """Returns a generator yielding immediate child types."""
    raise NotImplementedError

  @abc.abstractmethod
  def __repr__(self):
    """Returns a full-form representation of this type."""
    raise NotImplementedError

  def __str__(self):
    """Returns a concise representation of this type."""
    return self.compact_representation()

  @abc.abstractmethod
  def __hash__(self):
    """Produces a hash value for this type."""
    raise NotImplementedError

  @abc.abstractmethod
  def __eq__(self, other):
    """Determines whether two type definitions are identical.

    Note that this notion of equality is stronger than equivalence. Two types
    with equivalent definitions may not be identical, e.g., if they represent
    templates with differently named type variables in their definitions.

    Args:
      other: The other type to compare against.

    Returns:
      `True` if type definitions are syntactically identical (as defined above),
      otherwise `False`.
    """
    raise NotImplementedError

  def __ne__(self, other):
    return not self == other

  def check_assignable_from(self, source_type: 'Type') -> None:
    """Raises if values of `source_type` cannot be cast to this type."""
    if not self.is_assignable_from(source_type):
      raise TypeNotAssignableError(source_type=source_type, target_type=self)

  @abc.abstractmethod
  def is_assignable_from(self, source_type: 'Type') -> bool:
    """Returns whether values of `source_type` can be cast to this type."""
    raise NotImplementedError

  def check_equivalent_to(self, other: 'Type') -> None:
    """Raises if values of 'other' cannot be cast to and from this type."""
    if not self.is_equivalent_to(other):
      raise TypesNotEquivalentError(self, other)

  def is_equivalent_to(self, other: 'Type') -> bool:
    """Returns whether values of `other` can be cast to and from this type."""
    return self.is_assignable_from(other) and other.is_assignable_from(self)

  def is_identical_to(self, other: 'Type') -> bool:
    """Returns whether or not `self` and `other` are exactly identical."""
    return self == other


class _Intern(abc.ABCMeta):
  """A metaclass which interns instances.

  This is used to create classes where the following predicate holds:
  `MyClass(some_args) is MyClass(some_args)`

  That is, objects of the class with the same constructor parameters result
  in values with the same object identity. This can make comparison of deep
  structures much cheaper, since a shallow equality check can short-circuit
  comparison.

  Classes which set `_Intern` as a metaclass must have a
  `_hashable_from_init_args` classmethod which defines exactly the parameters
  passed to the `__init__` method. If one of the parameters passed to the
  `_Intern.__call__` is an iterator it will be converted to a list before
  `_hashable_from_init_args` and `__init__` are called.

  NOTE: also that this metaclass must only be used with *immutable* values, as
  mutation would cause all similarly-constructed instances to be mutated
  together.

  Inherits from `abc.ABCMeta` to prevent subclass conflicts.
  """

  @classmethod
  def _hashable_from_init_args(mcs, *args, **kwargs) -> Hashable:
    raise NotImplementedError

  def __call__(cls, *args, **kwargs):
    # Convert all `Iterator`s in both `args` and `kwargs` to `list`s so they can
    # be used in both `_hashable_from_init_args` and `__init__`.
    def _normalize(obj):
      if isinstance(obj, Iterator):
        return list(obj)
      else:
        return obj

    args = [_normalize(x) for x in args]
    kwargs = {k: _normalize(v) for k, v in kwargs.items()}

    # Salt the key with `cls` to account for two different classes that return
    # the same result from `_hashable_from_init_args`.
    key = (cls, cls._hashable_from_init_args(*args, **kwargs))
    intern_pool = _intern_pool[cls]
    instance = intern_pool.get(key, None)
    if instance is None:
      instance = super().__call__(*args, **kwargs)
      intern_pool[key] = instance
    return instance


# A per-`typing.Type` map from `__init__` arguments to object instances.
#
# This is used by the `_Intern` metaclass to allow reuse of object instances
# when new objects are requested with the same `__init__` arguments as
# existing object instances.
#
# Implementation note: this double-map is used rather than a single map
# stored as a field of each class because some class objects themselves would
# begin destruction before the map fields of other classes, causing errors
# during destruction.
_intern_pool: MutableMapping[type[Type], MutableMapping[Hashable, Type]] = (
    collections.defaultdict(dict)
)


def _clear_intern_pool() -> None:
  # We must clear our `WeakKeyValueDictionary`s at the end of the program to
  # prevent Python from deleting the standard library out from under us before
  # removing the  entries from the dictionary. Yes, this is cursed.
  #
  # If this isn't done, Python will call `__eq__` on our types after
  # `abc.ABCMeta` has already been deleted from the world, resulting in
  # exceptions after main.
  global _intern_pool
  _intern_pool = None


atexit.register(_clear_intern_pool)


_DtypeLike = Union[type[np.generic], np.dtype]


def _is_dtype_like(obj: object) -> TypeGuard[_DtypeLike]:
  """Returns `True` if `obj` is dtype like, otherwise `False`."""
  if isinstance(obj, type) and issubclass(obj, np.generic):
    return True
  else:
    return isinstance(obj, np.dtype)


def _is_array_shape_like(
    obj: object,
) -> TypeGuard[Union[array_shape._ArrayShapeLike]]:
  """Returns `True` if `obj` is an `_ArrayShapeLike`, otherwise `False`."""
  if obj is None:
    return True
  elif isinstance(obj, Sequence):
    # If iterating over the `Sequence` fails, then `obj` is not an
    # `array_shape._ArrayShapeLike`.
    try:
      return all(isinstance(x, int) or x is None for x in obj)
    except Exception:  # pylint: disable=broad-exception-caught
      return False
  else:
    return False


def _to_dtype(dtype: _DtypeLike) -> np.dtype:
  """Returns a `np.dtype` for the dtype like object.

  Normalize `dtype` to an instance of `np.dtype` that describes an array
  scalar. see https://numpy.org/doc/stable/reference/arrays.scalars.html.

  Args:
    dtype: A dtype like object.
  """
  if isinstance(dtype, np.dtype):
    dtype = dtype.type
  if dtype is np.bytes_:
    dtype = np.str_

  if not dtype_utils.is_valid_dtype(dtype):
    raise NotImplementedError(f'Unexpected dtype found: {dtype}.')
  return np.dtype(dtype)


class TensorType(Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing tensors."""

  @classmethod
  def _hashable_from_init_args(
      cls,
      dtype: _DtypeLike,
      shape: array_shape._ArrayShapeLike = (),
  ) -> Hashable:
    """Returns hashable `TensorType.__init__` args."""
    dtype = _to_dtype(dtype)
    if shape is not None:
      shape = tuple(shape)
    return (dtype, shape)

  def __init__(
      self,
      dtype: _DtypeLike,
      shape: array_shape._ArrayShapeLike = (),
  ):
    """Constructs a new instance from the given `dtype` and `shape`.

    Args:
      dtype: The `np.dtype` of the array.
      shape: The shape of the array.

    Raises:
      TypeError: if arguments are of the wrong types.
    """
    self._dtype = _to_dtype(dtype)
    if shape is not None:
      shape = tuple(shape)
    self._shape = shape
    self._proto = None

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'TensorType':
    """Returns a `TensorType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'tensor')

    dtype = dtype_utils.from_proto(type_pb.tensor.dtype)
    shape_pb = array_pb2.ArrayShape(
        dim=type_pb.tensor.dims,
        unknown_rank=type_pb.tensor.unknown_rank,
    )
    shape = array_shape.from_proto(shape_pb)
    return TensorType(dtype, shape)

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      dtype_pb = dtype_utils.to_proto(self.dtype.type)
      shape_pb = array_shape.to_proto(self.shape)
      tensor_type_pb = computation_pb2.TensorType(
          dtype=dtype_pb,
          dims=shape_pb.dim,
          unknown_rank=shape_pb.unknown_rank,
      )
      self._proto = computation_pb2.Type(tensor=tensor_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    return iter(())

  @property
  def dtype(self) -> np.dtype:
    return self._dtype

  @property
  def shape(self) -> array_shape.ArrayShape:
    return self._shape

  def __repr__(self):
    dtype_repr = f'np.{self._dtype.type.__name__}'
    if array_shape.is_shape_scalar(self._shape):
      return f'TensorType({dtype_repr})'
    else:
      return f'TensorType({dtype_repr}, {self._shape!r})'

  def __hash__(self):
    return hash((self._dtype, self._shape))

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, TensorType)
        and self._dtype == other.dtype
        and self._shape == other.shape
    )

  def is_assignable_from(self, source_type: Type) -> bool:
    if self is source_type:
      return True

    if (
        not isinstance(source_type, TensorType)
        or self.dtype != source_type.dtype
    ):
      return False

    target_shape = self.shape
    source_shape = source_type.shape
    if target_shape is None:
      return True
    elif source_shape is None:
      return False

    if len(target_shape) != len(source_shape):
      return False

    def _dimension_is_assignable_from(target_dim, source_dim):
      return (target_dim is None) or (target_dim == source_dim)

    return all(
        _dimension_is_assignable_from(x, y)
        for x, y in zip(target_shape, source_shape)
    )


def _format_struct_type_members(struct_type: 'StructType') -> str:
  def _element_repr(element):
    name, value = element
    if name is not None:
      return "('{}', {!r})".format(name, value)
    return repr(value)

  return ', '.join(_element_repr(e) for e in struct_type.items())


def _to_named_types(
    elements: Iterable[object],
) -> Sequence[tuple[Optional[str], Type]]:
  """Creates an `Iterable` of optionally named types from `elements`.

  This function creates an `Iterable` of optionally named types by iterating
  over `elements` and normalizing each element.

  If `elements` is an `Iterable` with named elements (e.g. `Mapping` or
  `NamedTuple`), the normalize element will have a name equal to the name of the
  element and a value equal to the value of the element convereted to a type
  using `to_type`.

  If `elements` is an `Iterable` with unnamed elements (e.g. list), the
  normalized element will have a name of `None` and a value equal to the element
  convereted to a type using `to_type`.

  NOTE: This function treats a single element being passed in as `elements` as
  if it were an iterable of that element.

  Args:
    elements: An iterable of named or unnamed objects to convert to
      `federated_language.Types`. See `federated_language.types.to_type` for
      more information.

  Returns:
    A `Sequence` where each each element is `tuple[Optional[str], Type]`.
  """

  if py_typecheck.is_name_value_pair(elements):
    elements = [elements]
  elif isinstance(elements, py_typecheck.SupportsNamedTuple):
    elements = elements._asdict().items()
  elif isinstance(elements, Mapping):
    elements = elements.items()

  def _to_named_value_pair(element: object) -> tuple[Optional[str], Type]:
    if py_typecheck.is_name_value_pair(element):
      name, value = element
    else:
      name = None
      value = element
    value = to_type(value)
    return (name, value)

  return [_to_named_value_pair(x) for x in elements]


def _reserved_names_in_elements(
    elements: Sequence[tuple[Optional[str], object]],
    reserved_names: Sequence[str],
) -> set[str]:
  element_names = {n for n, _ in elements if n is not None}
  return set(reserved_names).intersection(element_names)


class StructType(structure.Struct, Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing structures.

  Elements initialized by name can be accessed as `foo.name`, and otherwise by
  index, `foo[index]`.

  Elements can not be given names that would conflict with the methods and on
  this class.
  """

  @classmethod
  def _hashable_from_init_args(
      cls,
      elements: Iterable[object],
      *,
      convert: bool = True,
  ) -> Hashable:
    if convert:
      elements = _to_named_types(elements)
    invalid_names = _reserved_names_in_elements(elements, dir(cls))
    if invalid_names:
      raise ValueError(
          'Expected named elements to not match any reserved names, found'
          f' {invalid_names}.'
      )
    return (tuple(elements), convert)

  def __init__(
      self,
      elements: Iterable[object],
      *,
      convert: bool = True,
  ):
    """Constructs a new instance from the given element types.

    Args:
      elements: An iterable of element specifications. Each element
        specification is either a type spec (an instance of
        `federated_language.Type` or something convertible to it via
        `federated_language.types.to_type`) for the element, or a (name, spec)
        for elements that have defined names. Alternatively, one can supply here
        an instance of `collections.OrderedDict` mapping element names to their
        types (or things that are convertible to types).
      convert: A flag to determine if the elements should be converted using
        `federated_language.types.to_type` or not.
    """
    if convert:
      elements = _to_named_types(elements)
    structure.Struct.__init__(self, elements)

    self._proto = None

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'StructType':
    """Returns a `StructType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'struct')

    elements = []
    for element_pb in type_pb.struct.element:
      name = element_pb.name if element_pb.name else None
      element_type = Type.from_proto(element_pb.value)
      elements.append((name, element_type))
    return StructType(elements)

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      element_pbs = []
      for name, element in self.items():
        value_pb = element.to_proto()
        element_pb = computation_pb2.StructType.Element(
            name=name,
            value=value_pb,
        )
        element_pbs.append(element_pb)
      struct_type_pb = computation_pb2.StructType(element=element_pbs)
      self._proto = computation_pb2.Type(struct=struct_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    return (element for _, element in self.items())

  @property
  def python_container(self) -> Optional[type[object]]:
    return None

  def items(self) -> Iterator[tuple[Optional[str], Type]]:
    return structure.iter_elements(self)

  def fields(self) -> list[str]:
    return structure.name_list(self)

  def __repr__(self):
    members = _format_struct_type_members(self)
    return f'StructType([{members}])'

  def __hash__(self):
    # Salt to avoid overlap.
    return hash((structure.Struct.__hash__(self), 'NTT'))

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, StructType) and structure.Struct.__eq__(self, other)
    )

  def is_assignable_from(self, source_type: Type) -> bool:
    if self is source_type:
      return True
    if not isinstance(source_type, StructType):
      return False
    target_elements = list(self.items())
    source_elements = list(source_type.items())
    if len(target_elements) != len(source_elements):
      return False
    for (target_name, target_element), (source_name, source_element) in zip(
        target_elements, source_elements
    ):
      if source_name is not None and source_name != target_name:
        return False
      if not target_element.is_assignable_from(source_element):
        return False
    return True


class StructWithPythonType(StructType, metaclass=_Intern):
  """A representation of a structure paired with a Python container type.

  Elements can not be given names that would conflict with the methods and on
  this class.
  """

  @classmethod
  def _hashable_from_init_args(
      cls, elements: Iterable[object], container_type: type[object]
  ) -> Hashable:
    elements = _to_named_types(elements)
    invalid_names = _reserved_names_in_elements(elements, dir(cls))
    if invalid_names:
      raise ValueError(
          'Expected named elements to not match any reserved names, found'
          f' {invalid_names}.'
      )
    return (tuple(elements), container_type)

  def __init__(self, elements: Iterable[object], container_type: type[object]):
    super().__init__(elements)
    self._container_type = container_type

  @classmethod
  def from_proto(
      cls, type_pb: computation_pb2.Type, *, container_type: type[object]
  ) -> 'StructWithPythonType':
    """Returns a `StructWithPythonType` for the `type_pb`."""
    struct_type = super().from_proto(type_pb)
    return StructWithPythonType(struct_type.items(), container_type)

  @property
  def python_container(self) -> type[object]:
    return self._container_type

  def __repr__(self):
    members = _format_struct_type_members(self)
    return 'StructType([{}]) as {}'.format(
        members, self._container_type.__name__
    )

  def __hash__(self):
    # Salt to avoid overlap.
    return hash((structure.Struct.__hash__(self), 'NTTWPCT'))

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, StructWithPythonType)
        and (self._container_type == other._container_type)
        and structure.Struct.__eq__(self, other)
    )


class SequenceType(Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing sequences.

  IMPORTANT: since `SequenceType` is frequently backed by `tf.data.Dataset`
  which converts `list` to `tuple`, any `SequenceType` constructed with
  `StructWithPythonType` elements will convert any `list` python container type
  to `tuple` python container types for interoperability.
  """

  @classmethod
  def _hashable_from_init_args(cls, element: object) -> Hashable:
    element = to_type(element)
    return (element,)

  def __init__(self, element: object):
    """Constructs a new instance from the given `element` type.

    Args:
      element: A specification of the element type, either an instance of
        `federated_language.Type` or something convertible to it by
        `federated_language.types.to_type`.
    """

    def convert_struct_with_list_to_struct_with_tuple(type_spec: T) -> T:
      """Convert any StructWithPythonType using lists to use tuples."""
      # We ignore non-struct, non-tensor types, these are not well formed types
      # for sequence elements.
      if not isinstance(type_spec, StructType):
        return type_spec
      elements = [
          (name, convert_struct_with_list_to_struct_with_tuple(value))
          for name, value in type_spec.items()
      ]
      if not isinstance(type_spec, StructWithPythonType):
        return StructType(elements=elements)
      container_cls = type_spec.python_container
      return StructWithPythonType(
          elements=elements,
          container_type=tuple if container_cls is list else container_cls,
      )

    element = to_type(element)
    self._element = convert_struct_with_list_to_struct_with_tuple(element)
    self._proto = None

    children_types = _get_contained_children_types(self)
    if (
        children_types.federated
        or children_types.function
        or children_types.sequence
    ):
      raise ValueError(
          'Expected a `federated_language.SequenceType` to not contain'
          ' `federated_language.FederatedType`s,'
          ' `federated_language.FunctionType`s, or'
          f' `federated_language.SequenceType`s, found {self}.'
      )

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'SequenceType':
    """Returns a `SequenceType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'sequence')

    element_type = Type.from_proto(type_pb.sequence.element)
    return SequenceType(element_type)

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      element_pb = self._element.to_proto()
      sequence_type_pb = computation_pb2.SequenceType(element=element_pb)
      self._proto = computation_pb2.Type(sequence=sequence_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    yield self._element

  @property
  def element(self) -> Type:
    return self._element

  def __repr__(self):
    return 'SequenceType({!r})'.format(self._element)

  def __hash__(self):
    return hash(self._element)

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, SequenceType) and self._element == other.element
    )

  def is_assignable_from(self, source_type: Type) -> bool:
    if self is source_type:
      return True
    return isinstance(
        source_type, SequenceType
    ) and self.element.is_assignable_from(source_type.element)


class FunctionType(Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing functional types."""

  @classmethod
  def _hashable_from_init_args(
      cls, parameter: Optional[object], result: object
  ) -> Hashable:
    if parameter is not None:
      parameter = to_type(parameter)
    result = to_type(result)
    return (parameter, result)

  def __init__(self, parameter: Optional[object], result: object):
    """Constructs a new instance from the given `parameter` and `result` types.

    Args:
      parameter: A specification of the parameter type, either an instance of
        `federated_language.Type` or something convertible to it by
        `federated_language.types.to_type`. Multiple input arguments can be
        specified as a single `federated_language.StructType`.
      result: A specification of the result type, either an instance of
        `federated_language.Type` or something convertible to it by
        `federated_language.types.to_type`.
    """
    if parameter is not None:
      parameter = to_type(parameter)
    self._parameter = parameter
    self._result = to_type(result)
    self._proto = None

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'FunctionType':
    """Returns a `FunctionType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'function')

    if type_pb.function.HasField('parameter'):
      parameter_type = Type.from_proto(type_pb.function.parameter)
    else:
      parameter_type = None
    result_type = Type.from_proto(type_pb.function.result)
    return FunctionType(parameter_type, result_type)

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      if self._parameter is not None:
        parameter_pb = self._parameter.to_proto()
      else:
        parameter_pb = None
      result_pb = self._result.to_proto()
      function_type_pb = computation_pb2.FunctionType(
          parameter=parameter_pb, result=result_pb
      )
      self._proto = computation_pb2.Type(function=function_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    if self._parameter is not None:
      yield self._parameter
    yield self._result

  @property
  def parameter(self) -> Optional[Type]:
    return self._parameter

  @property
  def result(self) -> Type:
    return self._result

  def __repr__(self):
    return 'FunctionType({!r}, {!r})'.format(self._parameter, self._result)

  def __hash__(self):
    return hash((self._parameter, self._result))

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, FunctionType)
        and self._parameter == other.parameter
        and self._result == other.result
    )

  def is_assignable_from(self, source_type: Type) -> bool:
    if self is source_type:
      return True
    if not isinstance(source_type, FunctionType):
      return False
    if (self.parameter is None) != (source_type.parameter is None):
      return False
    # Note that function parameters are contravariant, so we invert the check.
    if (
        self.parameter is not None
        and source_type.parameter is not None
        and not source_type.parameter.is_assignable_from(self.parameter)
    ):
      return False
    return self.result.is_assignable_from(source_type.result)


class AbstractType(Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing abstract types."""

  @classmethod
  def _hashable_from_init_args(cls, label: str) -> Hashable:
    return (label,)

  def __init__(self, label: str):
    """Constructs a new instance from the given string `label`.

    Args:
      label: A string label of an abstract type. All occurrences of the label
        within a computation's type signature refer to the same concrete type.
    """
    self._label = label
    self._proto = None

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'AbstractType':
    """Returns a `AbstractType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'abstract')

    return AbstractType(type_pb.abstract.label)

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      abstract_type_pb = computation_pb2.AbstractType(label=self._label)
      self._proto = computation_pb2.Type(abstract=abstract_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    return iter(())

  @property
  def label(self) -> str:
    return self._label

  def __repr__(self):
    return "AbstractType('{}')".format(self._label)

  def __hash__(self):
    return hash(self._label)

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, AbstractType) and self._label == other.label
    )

  def is_assignable_from(self, source_type: Type) -> bool:
    del source_type  # Unused.
    raise TypeError('Abstract types are not comparable.')


class PlacementType(Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing the placement type.

  There is only one placement type, a language built-in, just as there is only
  one `int` or `str` type in Python. All instances of this class represent the
  same built-in language placement type.
  """

  @classmethod
  def _hashable_from_init_args(cls) -> Hashable:
    return ()

  def __init__(self):
    """Constructs a new instance."""
    self._proto = None

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'PlacementType':
    """Returns a `PlacementType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'placement')

    return PlacementType()

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      placement_type_pb = computation_pb2.PlacementType()
      self._proto = computation_pb2.Type(placement=placement_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    return iter(())

  def __repr__(self):
    return 'PlacementType()'

  def __hash__(self):
    return 0

  def __eq__(self, other):
    return (self is other) or isinstance(other, PlacementType)

  def is_assignable_from(self, source_type: Type) -> bool:
    if self is source_type:
      return True
    return isinstance(source_type, PlacementType)


class FederatedType(Type, metaclass=_Intern):
  """An implementation of `federated_language.Type` representing placed types."""

  @classmethod
  def _hashable_from_init_args(
      cls,
      member: object,
      placement: placements.PlacementLiteral,
      all_equal: Optional[bool] = None,
  ) -> Hashable:
    member = to_type(member)
    return (member, placement, all_equal)

  def __init__(
      self,
      member: object,
      placement: placements.PlacementLiteral,
      all_equal: Optional[bool] = None,
  ):
    """Constructs a new federated type instance.

    Args:
      member: An instance of `federated_language.Type` or something convertible
        to it, that represents the type of the member components of each value
        of this federated type.
      placement: The specification of placement that the member components of
        this federated type are hosted on. Must be either a placement literal
        such as `federated_language.SERVER` or `federated_language.CLIENTS` to
        refer to a globally defined placement, or a placement label to refer to
        a placement defined in other parts of a type signature. Specifying
        placement labels is not implemented yet.
      all_equal: A `bool` value that indicates whether all members of the
        federated type are equal (`True`), or are allowed to differ (`False`).
        If `all_equal` is `None`, the value is selected as the default for the
        placement, e.g., `True` for `federated_language.SERVER` and `False` for
        `federated_language.CLIENTS`.
    """
    self._member = to_type(member)
    self._placement = placement
    if all_equal is None:
      all_equal = placement.default_all_equal
    self._all_equal = all_equal
    self._proto = None

    children_types = _get_contained_children_types(self)
    if children_types.federated or children_types.function:
      raise ValueError(
          'Expected a `federated_language.FederatedType` to not contain'
          ' `federated_language.FederatedType`s or'
          f' `federated_language.FunctionType`s, found {self}.'
      )

  @classmethod
  def from_proto(cls, type_pb: computation_pb2.Type) -> 'FederatedType':
    """Returns a `FederatedType` for the `type_pb`."""
    _check_type_has_field(type_pb, 'federated')

    placement_oneof = type_pb.federated.placement.WhichOneof('placement')
    if placement_oneof == 'value':
      member_type = Type.from_proto(type_pb.federated.member)
      placement = placements.uri_to_placement_literal(
          type_pb.federated.placement.value.uri
      )
      return FederatedType(member_type, placement, type_pb.federated.all_equal)
    else:
      raise NotImplementedError(
          f'Unexpected placement found: {placement_oneof}.'
      )

  def to_proto(self) -> computation_pb2.Type:
    """Returns a `computation_pb2.Type` for this type."""
    if self._proto is None:
      placement_pb = computation_pb2.Placement(uri=self.placement.uri)
      placement_spec_pb = computation_pb2.PlacementSpec(value=placement_pb)
      member_pb = self._member.to_proto()
      federated_type_pb = computation_pb2.FederatedType(
          placement=placement_spec_pb,
          all_equal=self.all_equal,
          member=member_pb,
      )
      self._proto = computation_pb2.Type(federated=federated_type_pb)
    return self._proto

  def children(self) -> Iterator[Type]:
    yield self._member

  @property
  def member(self) -> Type:
    return self._member

  @property
  def placement(self) -> placements.PlacementLiteral:
    return self._placement

  @property
  def all_equal(self) -> bool:
    return self._all_equal

  def __repr__(self):
    return 'FederatedType({!r}, {!r}, {!r})'.format(
        self._member, self._placement, self._all_equal
    )

  def __hash__(self):
    return hash((self._member, self._placement, self._all_equal))

  def __eq__(self, other):
    return (self is other) or (
        isinstance(other, FederatedType)
        and self._member == other.member
        and self._placement == other.placement
        and self._all_equal == other.all_equal
    )

  def is_assignable_from(self, source_type: Type) -> bool:
    if self is source_type:
      return True
    return (
        isinstance(source_type, FederatedType)
        and self.member.is_assignable_from(source_type.member)
        and (not self.all_equal or source_type.all_equal)
        and self.placement is source_type.placement
    )


def to_type(obj: object) -> Type:
  """Converts the argument into an instance of `federated_language.Type`.

  Examples of arguments convertible to tensor types:

  ```python
  np.int32
  (np.int32, [10])
  (np.int32, [None])
  ```

  Examples of arguments convertible to flat named tuple types:

  ```python
  [np.int32, np.bool]
  (np.int32, np.bool)
  [('a', np.int32), ('b', np.bool)]
  ('a', np.int32)
  collections.OrderedDict([('a', np.int32), ('b', np.bool)])
  ```

  Examples of arguments convertible to nested named tuple types:

  ```python
  (np.int32, (np.float32, np.bool))
  (np.int32, (('x', np.float32), np.bool))
  ((np.int32, [1]), (('x', (np.float32, [2])), (np.bool, [3])))
  ```

  `attr.s` class instances can also be used to describe types by populating
  the fields with the corresponding types:

  ```python
  @attr.s(auto_attribs=True)
  class MyDataClass:
    int_scalar
    string_array

  obj = MyDataClass(...)
  type_spec = federated_language.types.to_type(obj)

  @federated_language.tensorflow.computation(type_spec)
  def work(my_data):
    assert isinstance(my_data, MyDataClass)
    ...
  ```

  Args:
    obj: Either an instance of `federated_language.Type`, or an argument
      convertible to `federated_language.Type`.

  Returns:
    An instance of `federated_language.Type` corresponding to the given `obj`.
  """
  if isinstance(obj, Type):
    return obj
  elif _is_dtype_like(obj):
    return TensorType(obj)  # pytype: disable=wrong-arg-types  # b/290661340
  elif (
      isinstance(obj, tuple)
      and len(obj) == 2
      and _is_dtype_like(obj[0])
      and _is_array_shape_like(obj[1])
  ):
    dtype, shape = obj
    return TensorType(dtype, shape)
  elif isinstance(obj, (list, tuple)):
    if any(py_typecheck.is_name_value_pair(e, name_type=str) for e in obj):
      # The sequence has a (name, value) elements, the whole sequence is most
      # likely intended to be a `Struct`, do not store the Python container.
      return StructType(obj)
    else:
      return StructWithPythonType(obj, type(obj))
  elif attrs.has(type(obj)):
    return StructWithPythonType(attrs.asdict(obj, recurse=False), type(obj))
  elif isinstance(obj, py_typecheck.SupportsNamedTuple):
    elements = [(k, np.dtype(v)) for k, v in obj.__annotations__.items()]
    return StructWithPythonType(elements, obj)
  elif isinstance(obj, Mapping):
    return StructWithPythonType(obj, type(obj))
  elif isinstance(obj, structure.Struct):
    return StructType(structure.to_elements(obj))
  else:
    raise TypeError(
        f'Unable to interpret an argument of type {type(obj)} as a type spec.'
    )


@attrs.define(frozen=True)
class _ContainedChildrenTypes:
  """The types of children `federated_language.Types` contained by a `federated_language.Type`.

  This data structure is used by `_get_contained_children_types` to package
  the types of children `federated_language.Types` contained by a
  `federated_language.Type` in a more
  convenient way.
  """

  tensor: bool = False
  struct: bool = False
  struct_with_python_type: bool = False
  sequence: bool = False
  function: bool = False
  abstract: bool = False
  placement: bool = False
  federated: bool = False


# Manual cache used rather than `cachetools.cached` due to incompatibility
# with `WeakValueDictionary`. We want to use a `WeakValueDictionary` so that
# cache entries are destroyed once the types they index no longer exist.
_contained_children_types_cache: MutableMapping[
    Type, _ContainedChildrenTypes
] = weakref.WeakValueDictionary({})


def _clear_contained_children_types_cache():
  # We must clear our `WeakKeyValueDictionary`s at the end of the program to
  # prevent Python from deleting the standard library out from under us before
  # removing the  entries from the dictionary. Yes, this is cursed.
  #
  # If this isn't done, Python will call `__eq__` on our types after
  # `abc.ABCMeta` has already been deleted from the world, resulting in
  # exceptions after main.
  global _contained_children_types_cache
  _contained_children_types_cache = None


atexit.register(_clear_contained_children_types_cache)


def _get_contained_children_types(type_spec: Type) -> _ContainedChildrenTypes:
  """Returns the types of children `federated_language.Types` contained by `type_spec`.

  The `_ContainedChildrenTypes` is cached so that this function can be used in
  performance sensitive operations.

  Args:
    type_spec: A `federated_language.Type`.

  Raises:
    RuntimeError: If the cache becomes corrupted in some unexpected way.
  """
  if _contained_children_types_cache is None:
    raise RuntimeError('Unexpected runtime error.')
  children_types = _contained_children_types_cache.get(type_spec, None)
  if children_types is not None:
    return children_types

  children_types = _ContainedChildrenTypes()
  for child_type in type_spec.children():
    # Create a mutable dict from the frozen `_ContainedChildrenTypes` instance;
    # add the child and grandchildren updates; and then evolve the instance.
    updates = attrs.asdict(children_types)
    if isinstance(child_type, TensorType):
      updates['tensor'] = True
    elif isinstance(child_type, StructType):
      updates['struct'] = True
    elif isinstance(child_type, StructWithPythonType):
      updates['struct_with_python_type'] = True
    elif isinstance(child_type, SequenceType):
      updates['sequence'] = True
    elif isinstance(child_type, FunctionType):
      updates['function'] = True
    elif isinstance(child_type, AbstractType):
      updates['abstract'] = True
    elif isinstance(child_type, PlacementType):
      updates['placement'] = True
    elif isinstance(child_type, FederatedType):
      updates['federated'] = True
    else:
      raise NotImplementedError(f'Unexpected type found: {type(child_type)}.')
    grandchildren_types = _get_contained_children_types(child_type)
    for key, value in attrs.asdict(grandchildren_types).items():
      if value:
        updates[key] = True
    children_types = attrs.evolve(children_types, **updates)

  _contained_children_types_cache[type_spec] = children_types
  return children_types


def _string_representation(type_spec: Type, formatted: bool) -> str:
  """Returns the string representation of a `Type`.

  This function creates a `list` of strings representing the given `type_spec`;
  combines the strings in either a formatted or un-formatted representation; and
  returns the resulting string representation.

  Args:
    type_spec: An instance of a `Type`.
    formatted: A boolean indicating if the returned string should be formatted.

  Raises:
    TypeError: If `type_spec` has an unexpected type.
  """

  def _combine(components):
    """Returns a `list` of strings by combining `components`.

    This function creates and returns a `list` of strings by combining a `list`
    of `components`. Each `component` is a `list` of strings representing a part
    of the string of a `Type`. The `components` are combined by iteratively
    **appending** the last element of the result with the first element of the
    `component` and then **extending** the result with remaining elements of the
    `component`.

    For example:

    >>> _combine([['a'], ['b'], ['c']])
    ['abc']

    >>> _combine([['a', 'b', 'c'], ['d', 'e', 'f']])
    ['abcd', 'ef']

    This function is used to help track where new-lines should be inserted into
    the string representation if the lines are formatted.

    Args:
      components: A `list` where each element is a `list` of strings
        representing a part of the string of a `Type`.
    """
    lines = ['']
    for component in components:
      lines[-1] = '{}{}'.format(lines[-1], component[0])
      lines.extend(component[1:])
    return lines

  def _indent(lines, indent_chars='  '):
    """Returns an indented `list` of strings."""
    return ['{}{}'.format(indent_chars, e) for e in lines]

  def _lines_for_named_types(named_type_specs, formatted):
    """Returns a `list` of strings representing the given `named_type_specs`.

    Args:
      named_type_specs: A `list` of named computations, each being a pair
        consisting of a name (either a string, or `None`) and a
        `ComputationBuildingBlock`.
      formatted: A boolean indicating if the returned string should be
        formatted.
    """
    lines = []
    for index, (name, type_spec) in enumerate(named_type_specs):
      if index != 0:
        if formatted:
          lines.append([',', ''])
        else:
          lines.append([','])
      element_lines = _lines_for_type(type_spec, formatted)
      if name is not None:
        element_lines = _combine([
            ['{}='.format(name)],
            element_lines,
        ])
      lines.append(element_lines)
    return _combine(lines)

  def _lines_for_type(type_spec, formatted):
    """Returns a `list` of strings representing the given `type_spec`.

    Args:
      type_spec: An instance of a `Type`.
      formatted: A boolean indicating if the returned string should be
        formatted.
    """
    if isinstance(type_spec, AbstractType):
      return [type_spec.label]
    elif isinstance(type_spec, FederatedType):
      member_lines = _lines_for_type(type_spec.member, formatted)
      placement_line = '@{}'.format(type_spec.placement)
      if type_spec.all_equal:
        return _combine([member_lines, [placement_line]])
      else:
        return _combine([['{'], member_lines, ['}'], [placement_line]])
    elif isinstance(type_spec, FunctionType):
      if type_spec.parameter is not None:
        parameter_lines = _lines_for_type(type_spec.parameter, formatted)
      else:
        parameter_lines = ['']
      result_lines = _lines_for_type(type_spec.result, formatted)
      return _combine([['('], parameter_lines, [' -> '], result_lines, [')']])
    elif isinstance(type_spec, StructType):
      if not type_spec:
        return ['<>']
      elements = list(type_spec.items())
      elements_lines = _lines_for_named_types(elements, formatted)
      if formatted:
        elements_lines = _indent(elements_lines)
        lines = [['<', ''], elements_lines, ['', '>']]
      else:
        lines = [['<'], elements_lines, ['>']]
      return _combine(lines)
    elif isinstance(type_spec, PlacementType):
      return ['placement']
    elif isinstance(type_spec, SequenceType):
      element_lines = _lines_for_type(type_spec.element, formatted)
      return _combine([element_lines, ['*']])
    elif isinstance(type_spec, TensorType):
      if type_spec.shape is None:
        return ['{!r}(shape=None)'.format(type_spec.dtype.name)]
      elif type_spec.shape:

        def _value_string(value):
          return str(value) if value is not None else '?'

        value_strings = [_value_string(e) for e in type_spec.shape]
        values_strings = ','.join(value_strings)
        return ['{}[{}]'.format(type_spec.dtype.name, values_strings)]
      else:
        return [type_spec.dtype.name]
    else:
      raise NotImplementedError(f'Unexpected type found: {type(type_spec)}.')

  lines = _lines_for_type(type_spec, formatted)
  lines = [line.rstrip() for line in lines]
  if formatted:
    return '\n'.join(lines)
  else:
    return ''.join(lines)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library of pure-python retrying decorators."""

import asyncio
from collections.abc import Callable
import functools
import inspect
import time
from typing import Union


def retry(
    fn=None,
    *,
    retry_on_exception_filter: Callable[[Exception], bool] = lambda x: True,
    retry_on_result_filter: Callable[[object], bool] = lambda x: False,
    wait_max_ms: Union[float, int] = 30000,
    wait_multiplier: Union[float, int] = 2,
):
  """Pure Python decorator that retries functions or coroutine functions.

  `retry` starts at some delay between function invocations, and backs
  off exponentialy with factor `wait_multiplier` until the max of
  `max_wait_ms`, at which point `retry` will continue to retry `fn` at intervals
  of `max_wait_ms` until `retry_on_exception_filter` returns `False`.

  Args:
    fn: Optional Python function or coroutine function to wrap in retrying
      logic. If None, `retry` will return a callable which decorates a function
      or corofunc to be passed later.
    retry_on_exception_filter: Function accepting a Python `Exception`, and
      returning a Boolean indicating whether or not to retry the invocation.
    retry_on_result_filter: Function accepting a function result or coroutine
      function result, and returning a Boolean indicating whether or not to
      retry the invocation.
    wait_max_ms: Maximum time `retry` is allowed to wait between invocations of
      `fn`, in milliseconds. Must be positive.
    wait_multiplier: Number determining the exponential backoff multiplier to
      use. Must be positive.

  Returns:
    In the case that `fn` is provided, a decorated version of `fn` respecting
    the semantics above. If `fn` is not provided, returns a callable which can
    be used to decorate a function or coroutine function at a later time.
  """
  if not inspect.isfunction(retry_on_exception_filter):
    raise TypeError(
        'Expected function to be passed as retry_on_exception_filter; '
        'encountered {} of type {}.'.format(
            retry_on_exception_filter, type(retry_on_exception_filter)
        )
    )
  if not inspect.isfunction(retry_on_result_filter):
    raise TypeError(
        'Expected function to be passed as retry_on_result_filter; '
        'encountered {} of type {}.'.format(
            retry_on_result_filter, type(retry_on_result_filter)
        )
    )
  if wait_max_ms <= 0:
    raise ValueError(
        'wait_max_ms required to be positive; encountered value {}.'.format(
            wait_max_ms
        )
    )
  if wait_multiplier <= 0:
    raise ValueError(
        'wait_multiplier required to be positive; encountered value {}.'.format(
            wait_multiplier
        )
    )

  if fn is None:
    # Called with arguments; delay decoration until `fn` is passed in.
    return functools.partial(
        retry,
        retry_on_exception_filter=retry_on_exception_filter,
        retry_on_result_filter=retry_on_result_filter,
        wait_max_ms=wait_max_ms,
        wait_multiplier=wait_multiplier,
    )

  if inspect.iscoroutinefunction(fn):
    # Similar to the logic in tracing.py, we case on corofunction versus vanilla
    # function.

    @functools.wraps(fn)
    async def retry_coro_fn(*args, **kwargs):
      retry_wait_ms = 1.0

      while True:
        try:
          result = await fn(*args, **kwargs)
          if retry_on_result_filter(result):
            retry_wait_ms = min(wait_max_ms, retry_wait_ms * wait_multiplier)
            # time.sleep takes arguments in seconds.
            await asyncio.sleep(retry_wait_ms / 1000)
            continue
          else:
            return result
        except Exception as e:  # pylint: disable=broad-except
          if not retry_on_exception_filter(e):
            raise e
          retry_wait_ms = min(wait_max_ms, retry_wait_ms * wait_multiplier)
          # asyncio.sleep takes arguments in seconds.
          await asyncio.sleep(retry_wait_ms / 1000)

    return retry_coro_fn

  elif inspect.isfunction(fn):
    # Vanilla Python function; decorate as normal.

    @functools.wraps(fn)
    def retry_fn(*args, **kwargs):
      retry_wait_ms = 1.0

      while True:
        try:
          result = fn(*args, **kwargs)
          if retry_on_result_filter(result):
            retry_wait_ms = min(wait_max_ms, retry_wait_ms * wait_multiplier)
            # time.sleep takes arguments in seconds.
            time.sleep(retry_wait_ms / 1000)
            continue
          else:
            return result
        except Exception as e:  # pylint: disable=broad-except
          if not retry_on_exception_filter(e):
            raise e
          retry_wait_ms = min(wait_max_ms, retry_wait_ms * wait_multiplier)
          # time.sleep takes arguments in seconds.
          time.sleep(retry_wait_ms / 1000)

    return retry_fn

  else:
    raise TypeError(
        'Retrying expects Python function or coroutine function; '
        'passed {} of type {}.'.format(fn, type(fn))
    )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines abstract interfaces for representing data sources."""

import abc
from typing import Optional

from federated_language.common_libs import serializable
from federated_language.types import computation_types


class FederatedDataSourceIterator(serializable.Serializable, abc.ABC):
  """An abstract interface for representing federated data source iterators.

  This interface abstracts away the specifics of iterating over data in a data
  source.

  Things one can do with a data source iterator:

  * Determine the type of the data supplied by this iterator by inspecting
    the `federated_type` property. The type returned must match that of the data
    source that returned this iterator.

  * Return a new selection of federated data from the iterator by invoking
    `select`.

  Please see `federated_language.program.FederatedDataSource` for additional
  context and the
  high-level description of how to use data sources.
  """

  @property
  @abc.abstractmethod
  def federated_type(self) -> computation_types.FederatedType:
    """The type of the data returned by calling `select`."""
    raise NotImplementedError

  @abc.abstractmethod
  def select(self, k: Optional[int] = None) -> object:
    """Returns a new selection of federated data from this iterator.

    Args:
      k: An optional number of elements to select. Must be a positive integer,
        or `None` if unspecified.

    Returns:
      An object of type `federated_type` representing the selected data, and
      that can be supplied as an argument to a `federated_language.Computation`.
      See
      `federated_language.program.FederatedContext` for more information about
      these types.
    """
    raise NotImplementedError


class FederatedDataSource(abc.ABC):
  """An abstract interface for representing federated data sources.

  This interface abstracts away the specifics of working with various types of
  data sources.

  Things one can do with a data source:

  * Determine the type of the data supplied by this data source by inspecting
    the `federated_type` property. The type returned should be a federated type.
    Note that depending on whether this data source contains one or a number of
    federated datasets, the type may or may not be a struct (with individual
    datasets appearing as elements of this struct).

  * Construct a new iterator for this data source by invoking `iterator` on it.
    Each iterator represents an independent pass over the data from this data
    source.
  """

  @property
  @abc.abstractmethod
  def federated_type(self) -> computation_types.FederatedType:
    """The type of the data returned by calling `select` on an iterator."""
    raise NotImplementedError

  @abc.abstractmethod
  def iterator(self) -> FederatedDataSourceIterator:
    """Returns a new iterator for retrieving data from this data source."""
    raise NotImplementedError



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: federated_language/proto/array.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'federated_language/proto/array.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from federated_language.proto import data_type_pb2 as federated__language_dot_proto_dot_data__type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$federated_language/proto/array.proto\x12\x12\x66\x65\x64\x65rated_language\x1a(federated_language/proto/data_type.proto\"/\n\nArrayShape\x12\x0b\n\x03\x64im\x18\x01 \x03(\x03\x12\x14\n\x0cunknown_rank\x18\x02 \x01(\x08\"\xa6\n\n\x05\x41rray\x12+\n\x05\x64type\x18\x01 \x01(\x0e\x32\x1c.federated_language.DataType\x12-\n\x05shape\x18\x02 \x01(\x0b\x32\x1e.federated_language.ArrayShape\x12\x14\n\x07\x63ontent\x18\x12 \x01(\x0cH\x01\x88\x01\x01\x12\x37\n\tbool_list\x18\x03 \x01(\x0b\x32\".federated_language.Array.BoolListH\x00\x12\x36\n\tint8_list\x18\x04 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x37\n\nint16_list\x18\x05 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x37\n\nint32_list\x18\x06 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x39\n\nint64_list\x18\x07 \x01(\x0b\x32#.federated_language.Array.Int64ListH\x00\x12\x37\n\nuint8_list\x18\x08 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x38\n\x0buint16_list\x18\t \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12;\n\x0buint32_list\x18\n \x01(\x0b\x32$.federated_language.Array.Uint32ListH\x00\x12;\n\x0buint64_list\x18\x0b \x01(\x0b\x32$.federated_language.Array.Uint64ListH\x00\x12\x39\n\x0c\x66loat16_list\x18\x0c \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12;\n\x0c\x66loat32_list\x18\r \x01(\x0b\x32#.federated_language.Array.FloatListH\x00\x12<\n\x0c\x66loat64_list\x18\x0e \x01(\x0b\x32$.federated_language.Array.DoubleListH\x00\x12=\n\x0e\x63omplex64_list\x18\x0f \x01(\x0b\x32#.federated_language.Array.FloatListH\x00\x12?\n\x0f\x63omplex128_list\x18\x10 \x01(\x0b\x32$.federated_language.Array.DoubleListH\x00\x12:\n\rbfloat16_list\x18\x13 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12:\n\x0bstring_list\x18\x11 \x01(\x0b\x32#.federated_language.Array.BytesListH\x00\x1a\x19\n\x08\x42oolList\x12\r\n\x05value\x18\x01 \x03(\x08\x1a\x18\n\x07IntList\x12\r\n\x05value\x18\x01 \x03(\x05\x1a\x1a\n\tInt64List\x12\r\n\x05value\x18\x01 \x03(\x03\x1a\x1b\n\nUint32List\x12\r\n\x05value\x18\x01 \x03(\r\x1a\x1b\n\nUint64List\x12\r\n\x05value\x18\x01 \x03(\x04\x1a\x1a\n\tFloatList\x12\r\n\x05value\x18\x01 \x03(\x02\x1a\x1b\n\nDoubleList\x12\r\n\x05value\x18\x01 \x03(\x01\x1a\x1a\n\tBytesList\x12\r\n\x05value\x18\x01 \x03(\x0c\x42\x06\n\x04kindB\n\n\x08_contentb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'federated_language.proto.array_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ARRAYSHAPE']._serialized_start=102
  _globals['_ARRAYSHAPE']._serialized_end=149
  _globals['_ARRAY']._serialized_start=152
  _globals['_ARRAY']._serialized_end=1470
  _globals['_ARRAY_BOOLLIST']._serialized_start=1228
  _globals['_ARRAY_BOOLLIST']._serialized_end=1253
  _globals['_ARRAY_INTLIST']._serialized_start=1255
  _globals['_ARRAY_INTLIST']._serialized_end=1279
  _globals['_ARRAY_INT64LIST']._serialized_start=1281
  _globals['_ARRAY_INT64LIST']._serialized_end=1307
  _globals['_ARRAY_UINT32LIST']._serialized_start=1309
  _globals['_ARRAY_UINT32LIST']._serialized_end=1336
  _globals['_ARRAY_UINT64LIST']._serialized_start=1338
  _globals['_ARRAY_UINT64LIST']._serialized_end=1365
  _globals['_ARRAY_FLOATLIST']._serialized_start=1367
  _globals['_ARRAY_FLOATLIST']._serialized_end=1393
  _globals['_ARRAY_DOUBLELIST']._serialized_start=1395
  _globals['_ARRAY_DOUBLELIST']._serialized_end=1422
  _globals['_ARRAY_BYTESLIST']._serialized_start=1424
  _globals['_ARRAY_BYTESLIST']._serialized_end=1450
# @@protoc_insertion_point(module_scope)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library of construction functions for building block structures."""

from collections.abc import Iterator, Sequence
import functools
import random
import string
from typing import Optional, Union

from federated_language.compiler import building_blocks
from federated_language.compiler import intrinsic_defs
from federated_language.compiler import transformation_utils
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_transformations

Index = Union[str, int]
Path = Union[Index, tuple[Index, ...]]


def select_output_from_lambda(
    comp: building_blocks.Lambda, paths: Union[Path, list[Path]]
) -> building_blocks.Lambda:
  """Constructs a new function with result of selecting `paths` from `comp`.

  Args:
    comp: Lambda computation with result type `federated_language.StructType`
      from which we wish to select the sub-results at `paths`.
    paths: Either a `Path` or list of `Path`s specifying the indices we wish to
      select from the result of `comp`. Each path must be a `tuple` of `str` or
      `int` indices from which to select an output. If `paths` is a list, the
      returned computation will have a `federated_language.StructType` result
      holding each of the specified selections.

  Returns:
    A version of `comp` with result value the selection from the result of
    `comp` specified by `paths`.
  """
  if not isinstance(comp.type_signature.result, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {comp.type_signature.result}.'
    )

  def _select_path(result, path: Path):
    if not isinstance(path, tuple):
      path = (path,)
    for index in path:
      if isinstance(result, building_blocks.Struct):
        result = result[index]
      elif isinstance(index, str):
        result = building_blocks.Selection(result, name=index)
      elif isinstance(index, int):
        result = building_blocks.Selection(result, index=index)
      else:
        raise TypeError(
            'Invalid selection type: expected `str` or `int`, '
            f'found value `{index}` of type `{type(index)}`.'
        )
    return result

  if isinstance(paths, list):
    # Avoid duplicating `comp.result` by binding it to a local.
    result_name = next(unique_name_generator(comp))
    result_ref = building_blocks.Reference(
        result_name, comp.result.type_signature
    )
    elements = [_select_path(result_ref, path) for path in paths]
    result = building_blocks.Block(
        [(result_name, comp.result)], building_blocks.Struct(elements)
    )
  else:
    result = _select_path(comp.result, paths)
  return building_blocks.Lambda(
      comp.parameter_name, comp.parameter_type, result
  )


def unique_name_generator(
    comp: building_blocks.ComputationBuildingBlock, prefix: str = '_var'
) -> Iterator[str]:
  """Yields a new unique name that does not exist in `comp`.

  Args:
    comp: The computation building block to use as a reference.
    prefix: The prefix to use when generating unique names. If `prefix` is
      `None` or if `comp` contains any name with this prefix, then a unique
      prefix will be generated from random lowercase ascii characters.
  """
  if comp is not None:
    names = transformation_utils.get_unique_names(comp)
  else:
    names = set()
  while prefix is None or any(n.startswith(prefix) for n in names):
    characters = string.ascii_lowercase
    prefix = '_{}'.format(''.join(random.choice(characters) for _ in range(3)))
  index = 1
  while True:
    yield '{}{}'.format(prefix, index)
    index += 1


@functools.lru_cache()
def create_identity(
    type_signature: computation_types.Type,
) -> building_blocks.Lambda:
  return building_blocks.Lambda(
      'id_arg',
      type_signature,
      building_blocks.Reference('id_arg', type_signature),
  )


class SelectionSpec:
  """Data class representing map from input tuple to selection of result.

  Attributes:
    tuple_index: The index of the source of the selection sequence in the
      desired result of the generated TensorFlow. If this `SelectionSpec`
      appears at index i of a list of `SelectionSpec`s, index j is the source
      for the result of the generated function at index i.
    selection_sequence: A list or tuple representing the selections to make from
      `tuple_index`, so that the list `[0]` for example would represent the
      output is the 0th element of `tuple_index`, while `[0, 0]` would represent
      that the output is the 0th element of the 0th element of `tuple_index`.
  """

  def __init__(self, tuple_index: int, selection_sequence: Sequence[int]):
    self._tuple_index = tuple_index
    self._selection_sequence = selection_sequence

  @property
  def tuple_index(self):
    return self._tuple_index

  @property
  def selection_sequence(self):
    return self._selection_sequence

  def __str__(self):
    return 'SelectionSequence(tuple_index={},selection_sequence={}'.format(
        self._tuple_index, self._selection_sequence
    )

  def __repr__(self):
    return str(self)


def create_federated_getitem_call(
    arg: building_blocks.ComputationBuildingBlock, idx: Union[int, slice]
) -> building_blocks.Call:
  """Creates computation building block passing getitem to federated value.

  Args:
    arg: Instance of `building_blocks.ComputationBuildingBlock` of
      `computation_types.FederatedType` with member of type
      `computation_types.StructType` from which we wish to pick out item `idx`.
    idx: Index, instance of `int` or `slice` used to address the
      `computation_types.StructType` underlying `arg`.

  Returns:
    Returns a `building_blocks.Call` with type signature
    `computation_types.FederatedType` of same placement as `arg`, the result
    of applying or mapping the appropriate `__getitem__` function, as defined
    by `idx`.
  """
  if not isinstance(arg.type_signature, computation_types.FederatedType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {arg.type_signature}.'
    )
  if not isinstance(arg.type_signature.member, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {arg.type_signature.member}.'
    )

  getitem_comp = create_federated_getitem_comp(arg, idx)
  return create_federated_map_or_apply(getitem_comp, arg)


def create_federated_getattr_call(
    arg: building_blocks.ComputationBuildingBlock, name: str
) -> building_blocks.Call:
  """Creates computation building block passing getattr to federated value.

  Args:
    arg: Instance of `building_blocks.ComputationBuildingBlock` of
      `computation_types.FederatedType` with member of type
      `computation_types.StructType` from which we wish to pick out item `name`.
    name: String name to address the `computation_types.StructType` underlying
      `arg`.

  Returns:
    Returns a `building_blocks.Call` with type signature
    `computation_types.FederatedType` of same placement as `arg`,
    the result of applying or mapping the appropriate `__getattr__` function,
    as defined by `name`.
  """
  if not isinstance(arg.type_signature, computation_types.FederatedType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {arg.type_signature}.'
    )
  if not isinstance(arg.type_signature.member, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {arg.type_signature.member}.'
    )

  getattr_comp = create_federated_getattr_comp(arg, name)
  return create_federated_map_or_apply(getattr_comp, arg)


def create_federated_getattr_comp(
    comp: building_blocks.ComputationBuildingBlock, name: str
) -> building_blocks.Lambda:
  """Function to construct computation for `federated_apply` of `__getattr__`.

  Creates a `building_blocks.ComputationBuildingBlock`
  which selects `name` from its argument, of type `comp.type_signature.member`,
  an instance of `computation_types.StructType`.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock` with type
      signature `computation_types.FederatedType` whose `member` attribute is of
      type `computation_types.StructType`.
    name: String name of attribute to grab.

  Returns:
    Instance of `building_blocks.Lambda` which grabs attribute
      according to `name` of its argument.
  """
  if not isinstance(comp.type_signature, computation_types.FederatedType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {comp.type_signature}.'
    )
  if not isinstance(comp.type_signature.member, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {comp.type_signature.member}.'
    )

  element_names = [
      x for x, _ in comp.type_signature.member.items()  # pytype: disable=attribute-error
  ]
  if name not in element_names:
    raise ValueError(
        'The federated value has no element of name `{}`. Value: {}'.format(
            name, comp.formatted_representation()
        )
    )
  apply_input = building_blocks.Reference('x', comp.type_signature.member)  # pytype: disable=attribute-error
  selected = building_blocks.Selection(apply_input, name=name)
  apply_lambda = building_blocks.Lambda(
      'x', apply_input.type_signature, selected
  )
  return apply_lambda


def create_federated_getitem_comp(
    comp: building_blocks.ComputationBuildingBlock, key: Union[int, slice]
) -> building_blocks.Lambda:
  """Function to construct computation for `federated_apply` of `__getitem__`.

  Creates a `building_blocks.ComputationBuildingBlock`
  which selects `key` from its argument, of type `comp.type_signature.member`,
  of type `computation_types.StructType`.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock` with type
      signature `computation_types.FederatedType` whose `member` attribute is of
      type `computation_types.StructType`.
    key: Instance of `int` or `slice`, key used to grab elements from the member
      of `comp`. implementation of slicing for `Value` objects with
      `type_signature` `computation_types.StructType`.

  Returns:
    Instance of `building_blocks.Lambda` which grabs slice
      according to `key` of its argument.
  """
  if not isinstance(comp.type_signature, computation_types.FederatedType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {comp.type_signature}.'
    )
  if not isinstance(comp.type_signature.member, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {comp.type_signature.member}.'
    )

  apply_input = building_blocks.Reference('x', comp.type_signature.member)  # pytype: disable=attribute-error
  if isinstance(key, int):
    selected = building_blocks.Selection(apply_input, index=key)
  else:
    elems = list(comp.type_signature.member.items())  # pytype: disable=attribute-error
    index_range = range(*key.indices(len(elems)))
    elem_list = []
    for k in index_range:
      elem_list.append(
          (elems[k][0], building_blocks.Selection(apply_input, index=k))
      )
    selected = building_blocks.Struct(elem_list)
  apply_lambda = building_blocks.Lambda(
      'x', apply_input.type_signature, selected
  )
  return apply_lambda


def _unname_fn_parameter(fn, unnamed_parameter_type):
  """Coerces `fn` to a comp whose parameter type is `unnamed_parameter_type`."""
  if any([n for n, _ in fn.type_signature.parameter.items()]):  # pytype: disable=attribute-error
    return building_blocks.Lambda(
        'a',
        unnamed_parameter_type,
        building_blocks.Call(
            fn,
            building_blocks.Reference('a', unnamed_parameter_type),
        ),
    )
  else:
    return fn


def create_federated_aggregate(
    value: building_blocks.ComputationBuildingBlock,
    zero: building_blocks.ComputationBuildingBlock,
    accumulate: building_blocks.ComputationBuildingBlock,
    merge: building_blocks.ComputationBuildingBlock,
    report: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated aggregate.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp, Comp, Comp, Comp]

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.
    zero: A `building_blocks.ComputationBuildingBlock` to use as the initial
      value.
    accumulate: A `building_blocks.ComputationBuildingBlock` to use as the
      accumulate function.
    merge: A `building_blocks.ComputationBuildingBlock` to use as the merge
      function.
    report: A `building_blocks.ComputationBuildingBlock` to use as the report
      function.

  Returns:
    A `building_blocks.Call`.
  """
  # Its okay if the first argument of accumulate is assignable from the zero,
  # without being the exact type. This occurs when accumulate has a type like
  # (<int32[?], int32> -> int32[?]) but zero is int32[0].
  zero_arg_type = accumulate.type_signature.parameter[0]  # pytype: disable=attribute-error
  zero_arg_type.check_assignable_from(zero.type_signature)
  result_type = computation_types.FederatedType(
      report.type_signature.result,  # pytype: disable=attribute-error
      placements.SERVER,
  )

  accumulate_parameter_type = computation_types.StructType([
      zero_arg_type,
      value.type_signature.member,  # pytype: disable=attribute-error
  ])
  accumulate = _unname_fn_parameter(accumulate, accumulate_parameter_type)
  merge_parameter_type = computation_types.StructType(
      [zero_arg_type, zero_arg_type]
  )
  merge = _unname_fn_parameter(merge, merge_parameter_type)

  intrinsic_type = computation_types.FunctionType(
      (
          computation_types.FederatedType(
              value.type_signature.member,  # pytype: disable=attribute-error
              value.type_signature.placement,  # pytype: disable=attribute-error
              all_equal=False,
          ),
          zero_arg_type,
          accumulate.type_signature,
          merge.type_signature,
          report.type_signature,
      ),
      result_type,
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_AGGREGATE.uri, intrinsic_type
  )
  values = building_blocks.Struct((value, zero, accumulate, merge, report))
  return building_blocks.Call(intrinsic, values)


def create_federated_apply(
    fn: building_blocks.ComputationBuildingBlock,
    arg: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated apply.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp]

  Args:
    fn: A `building_blocks.ComputationBuildingBlock` to use as the function.
    arg: A `building_blocks.ComputationBuildingBlock` to use as the argument.

  Returns:
    A `building_blocks.Call`.
  """
  result_type = computation_types.FederatedType(
      fn.type_signature.result,  # pytype: disable=attribute-error
      placements.SERVER,
  )
  intrinsic_type = computation_types.FunctionType(
      (fn.type_signature, arg.type_signature), result_type
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_APPLY.uri, intrinsic_type
  )
  values = building_blocks.Struct((fn, arg))
  return building_blocks.Call(intrinsic, values)


def create_federated_broadcast(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated broadcast.

            Call
           /    \
  Intrinsic      Comp

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.

  Returns:
    A `building_blocks.Call`.
  """
  result_type = computation_types.FederatedType(
      value.type_signature.member,  # pytype: disable=attribute-error
      placements.CLIENTS,
      all_equal=True,
  )
  intrinsic_type = computation_types.FunctionType(
      value.type_signature, result_type
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_BROADCAST.uri, intrinsic_type
  )
  return building_blocks.Call(intrinsic, value)


def create_federated_eval(
    fn: building_blocks.ComputationBuildingBlock,
    placement: placements.PlacementLiteral,
) -> building_blocks.Call:
  r"""Creates a called federated eval.

            Call
           /    \
  Intrinsic      Comp

  Args:
    fn: A `building_blocks.ComputationBuildingBlock` to use as the function.
    placement: A `placements.PlacementLiteral` to use as the placement.

  Returns:
    A `building_blocks.Call`.

  Raises:
    TypeError: If any of the types do not match.
  """
  if not isinstance(fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {fn.type_signature}.'
    )

  if placement is placements.CLIENTS:
    uri = intrinsic_defs.FEDERATED_EVAL_AT_CLIENTS.uri
    all_equal = False
  elif placement is placements.SERVER:
    uri = intrinsic_defs.FEDERATED_EVAL_AT_SERVER.uri
    all_equal = True
  else:
    raise TypeError('Unsupported placement {}.'.format(placement))
  result_type = computation_types.FederatedType(
      fn.type_signature.result,  # pytype: disable=attribute-error
      placement,
      all_equal=all_equal,
  )
  intrinsic_type = computation_types.FunctionType(
      fn.type_signature, result_type
  )
  intrinsic = building_blocks.Intrinsic(uri, intrinsic_type)
  return building_blocks.Call(intrinsic, fn)


def create_federated_map(
    fn: building_blocks.ComputationBuildingBlock,
    arg: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated map.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp]

  Args:
    fn: A `building_blocks.ComputationBuildingBlock` to use as the function.
    arg: A `building_blocks.ComputationBuildingBlock` to use as the argument.

  Returns:
    A `building_blocks.Call`.
  """
  parameter_type = computation_types.FederatedType(
      arg.type_signature.member,  # pytype: disable=attribute-error
      placements.CLIENTS,
  )
  result_type = computation_types.FederatedType(
      fn.type_signature.result,  # pytype: disable=attribute-error
      placements.CLIENTS,
  )
  intrinsic_type = computation_types.FunctionType(
      (fn.type_signature, parameter_type), result_type
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_MAP.uri, intrinsic_type
  )
  values = building_blocks.Struct((fn, arg))
  return building_blocks.Call(intrinsic, values)


def create_federated_map_all_equal(
    fn: building_blocks.ComputationBuildingBlock,
    arg: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated map of equal values.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp]

  NOTE: The `fn` is required to be deterministic and therefore should contain no
  `building_blocks.CompiledComputations`.

  Args:
    fn: A `building_blocks.ComputationBuildingBlock` to use as the function.
    arg: A `building_blocks.ComputationBuildingBlock` to use as the argument.

  Returns:
    A `building_blocks.Call`.
  """
  parameter_type = computation_types.FederatedType(
      arg.type_signature.member,  # pytype: disable=attribute-error
      placements.CLIENTS,
      all_equal=True,
  )
  result_type = computation_types.FederatedType(
      fn.type_signature.result,  # pytype: disable=attribute-error
      placements.CLIENTS,
      all_equal=True,
  )
  intrinsic_type = computation_types.FunctionType(
      (fn.type_signature, parameter_type), result_type
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_MAP_ALL_EQUAL.uri, intrinsic_type
  )
  values = building_blocks.Struct((fn, arg))
  return building_blocks.Call(intrinsic, values)


def create_federated_map_or_apply(
    fn: building_blocks.ComputationBuildingBlock,
    arg: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated map or apply depending on `arg`s placement.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp]

  Args:
    fn: A `building_blocks.ComputationBuildingBlock` to use as the function.
    arg: A `building_blocks.ComputationBuildingBlock` to use as the argument.

  Returns:
    A `building_blocks.Call`.
  """
  if arg.type_signature.placement is placements.CLIENTS:  # pytype: disable=attribute-error
    if arg.type_signature.all_equal:  # pytype: disable=attribute-error
      return create_federated_map_all_equal(fn, arg)
    else:
      return create_federated_map(fn, arg)
  elif arg.type_signature.placement is placements.SERVER:  # pytype: disable=attribute-error
    return create_federated_apply(fn, arg)
  else:
    raise TypeError(
        'Unsupported placement {}.'.format(arg.type_signature.placement)  # pytype: disable=attribute-error
    )


def create_federated_mean(
    value: building_blocks.ComputationBuildingBlock,
    weight: Optional[building_blocks.ComputationBuildingBlock],
) -> building_blocks.Call:
  r"""Creates a called federated mean.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp]

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.
    weight: A `building_blocks.ComputationBuildingBlock` to use as the weight or
      `None`.

  Returns:
    A `building_blocks.Call`.

  Raises:
    TypeError: If any of the types do not match.
  """
  result_type = computation_types.FederatedType(
      value.type_signature.member,  # pytype: disable=attribute-error
      placements.SERVER,
  )
  if weight is not None:
    intrinsic_type = computation_types.FunctionType(
        (
            computation_types.FederatedType(
                value.type_signature.member,  # pytype: disable=attribute-error
                value.type_signature.placement,  # pytype: disable=attribute-error
                all_equal=False,
            ),
            computation_types.FederatedType(
                weight.type_signature.member,  # pytype: disable=attribute-error
                weight.type_signature.placement,  # pytype: disable=attribute-error
                all_equal=False,
            ),
        ),
        result_type,
    )
    intrinsic = building_blocks.Intrinsic(
        intrinsic_defs.FEDERATED_WEIGHTED_MEAN.uri, intrinsic_type
    )
    values = building_blocks.Struct((value, weight))
    return building_blocks.Call(intrinsic, values)
  else:
    intrinsic_type = computation_types.FunctionType(
        computation_types.FederatedType(
            value.type_signature.member,  # pytype: disable=attribute-error
            value.type_signature.placement,  # pytype: disable=attribute-error
            all_equal=False,
        ),
        result_type,
    )
    intrinsic = building_blocks.Intrinsic(
        intrinsic_defs.FEDERATED_MEAN.uri, intrinsic_type
    )
    return building_blocks.Call(intrinsic, value)


def create_federated_min(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated min.

            Call
           /    \
  Intrinsic      Comp

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.

  Returns:
    A `building_blocks.Call`.

  Raises:
    ValueError: If any of the types do not match.
  """
  if not isinstance(value.type_signature, computation_types.FederatedType):
    raise ValueError('Expected a federated value.')
  result_type = computation_types.FederatedType(
      value.type_signature.member,
      placements.SERVER,
  )
  intrinsic_type = computation_types.FunctionType(
      computation_types.FederatedType(
          value.type_signature.member,  # pytype: disable=attribute-error
          value.type_signature.placement,  # pytype: disable=attribute-error
          all_equal=False,
      ),
      result_type,
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_MIN.uri, intrinsic_type
  )
  return building_blocks.Call(intrinsic, value)


def create_federated_max(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated max.

            Call
           /    \
  Intrinsic      Comp

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.

  Returns:
    A `building_blocks.Call`.

  Raises:
    ValueError: If any of the types do not match.
  """
  if not isinstance(value.type_signature, computation_types.FederatedType):
    raise ValueError('Expected a federated value.')
  result_type = computation_types.FederatedType(
      value.type_signature.member,
      placements.SERVER,
  )
  intrinsic_type = computation_types.FunctionType(
      computation_types.FederatedType(
          value.type_signature.member,  # pytype: disable=attribute-error
          value.type_signature.placement,  # pytype: disable=attribute-error
          all_equal=False,
      ),
      result_type,
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_MAX.uri, intrinsic_type
  )
  return building_blocks.Call(intrinsic, value)


def create_federated_secure_sum(
    value: building_blocks.ComputationBuildingBlock,
    max_input: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called secure sum.

            Call
           /    \
  Intrinsic      [Comp, Comp]

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.
    max_input: A `building_blocks.ComputationBuildingBlock` to use as the
      `max_input` value.

  Returns:
    A `building_blocks.Call`.
  """
  result_type = computation_types.FederatedType(
      value.type_signature.member,  # pytype: disable=attribute-error
      placements.SERVER,
  )
  intrinsic_type = computation_types.FunctionType(
      [
          computation_types.FederatedType(
              value.type_signature.member,  # pytype: disable=attribute-error
              value.type_signature.placement,  # pytype: disable=attribute-error
              all_equal=False,
          ),
          max_input.type_signature,
      ],
      result_type,
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_SECURE_SUM.uri, intrinsic_type
  )
  values = building_blocks.Struct([value, max_input])
  return building_blocks.Call(intrinsic, values)


def create_federated_secure_sum_bitwidth(
    value: building_blocks.ComputationBuildingBlock,
    bitwidth: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called secure sum using bitwidth.

            Call
           /    \
  Intrinsic      [Comp, Comp]

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.
    bitwidth: A `building_blocks.ComputationBuildingBlock` to use as the
      bitwidth value.

  Returns:
    A `building_blocks.Call`.
  """
  result_type = computation_types.FederatedType(
      value.type_signature.member,  # pytype: disable=attribute-error
      placements.SERVER,
  )
  intrinsic_type = computation_types.FunctionType(
      [
          computation_types.FederatedType(
              value.type_signature.member,  # pytype: disable=attribute-error
              value.type_signature.placement,  # pytype: disable=attribute-error
              all_equal=False,
          ),
          bitwidth.type_signature,
      ],
      result_type,
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_SECURE_SUM_BITWIDTH.uri, intrinsic_type
  )
  values = building_blocks.Struct([value, bitwidth])
  return building_blocks.Call(intrinsic, values)


def create_federated_select(
    client_keys: building_blocks.ComputationBuildingBlock,
    max_key: building_blocks.ComputationBuildingBlock,
    server_val: building_blocks.ComputationBuildingBlock,
    select_fn: building_blocks.ComputationBuildingBlock,
    secure: bool,
) -> building_blocks.Call:
  """Creates a called `federated_select` or `federated_secure_select`."""
  if not isinstance(max_key.type_signature, computation_types.FederatedType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {max_key.type_signature}.'
    )
  if not isinstance(server_val.type_signature, computation_types.FederatedType):
    raise ValueError(
        'Expected a `federated_language.FederatedType`, found'
        f' {server_val.type_signature}.'
    )
  if not isinstance(select_fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected a `federated_language.FunctionType`, found'
        f' {select_fn.type_signature}.'
    )

  single_key_type = max_key.type_signature.member
  select_fn_unnamed_param_type = computation_types.StructType([
      (None, server_val.type_signature.member),
      (None, single_key_type),
  ])
  select_fn = _unname_fn_parameter(select_fn, select_fn_unnamed_param_type)
  result_type = computation_types.FederatedType(
      computation_types.SequenceType(select_fn.type_signature.result),
      placements.CLIENTS,
  )
  intrinsic_type = computation_types.FunctionType(
      [
          computation_types.FederatedType(
              client_keys.type_signature.member,  # pytype: disable=attribute-error
              client_keys.type_signature.placement,  # pytype: disable=attribute-error
              all_equal=False,
          ),
          max_key.type_signature,
          server_val.type_signature,
          select_fn.type_signature,
      ],
      result_type,
  )
  if secure:
    intrinsic_def = intrinsic_defs.FEDERATED_SECURE_SELECT
  else:
    intrinsic_def = intrinsic_defs.FEDERATED_SELECT
  intrinsic = building_blocks.Intrinsic(intrinsic_def.uri, intrinsic_type)
  values = building_blocks.Struct([client_keys, max_key, server_val, select_fn])
  return building_blocks.Call(intrinsic, values)


def create_federated_sum(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated sum.

            Call
           /    \
  Intrinsic      Comp

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.

  Returns:
    A `building_blocks.Call`.
  """
  result_type = computation_types.FederatedType(
      value.type_signature.member,  # pytype: disable=attribute-error
      placements.SERVER,
  )
  intrinsic_type = computation_types.FunctionType(
      computation_types.FederatedType(
          value.type_signature.member,  # pytype: disable=attribute-error
          value.type_signature.placement,  # pytype: disable=attribute-error
          all_equal=False,
      ),
      result_type,
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.FEDERATED_SUM.uri, intrinsic_type
  )
  return building_blocks.Call(intrinsic, value)


def create_federated_unzip(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Block:
  r"""Creates a tuple of called federated maps or applies.

                Block
               /     \
  [value=Comp]        Tuple
                      |
                      [Call,                        Call, ...]
                      /    \                       /    \
             Intrinsic      Tuple         Intrinsic      Tuple
                            |                            |
                [Lambda(arg), Ref(value)]    [Lambda(arg), Ref(value)]
                            \                            \
                             Sel(0)                       Sel(1)
                                   \                            \
                                    Ref(arg)                     Ref(arg)

  This function returns a tuple of federated values given a `value` with a
  federated tuple type signature.

  Args:
    value: A `building_blocks.ComputationBuildingBlock` with a `type_signature`
      of type `computation_types.StructType` containing at least one element.

  Returns:
    A `building_blocks.Block`.

  Raises:
    ValueError: If `value` does not contain any elements.
  """
  named_type_signatures = list(value.type_signature.member.items())  # pytype: disable=attribute-error
  length = len(named_type_signatures)
  if length == 0:
    raise ValueError('federated_zip is only supported on non-empty tuples.')
  value_ref = building_blocks.Reference('value', value.type_signature)
  elements = []
  fn_ref = building_blocks.Reference('arg', named_type_signatures)
  for index, (name, _) in enumerate(named_type_signatures):
    sel = building_blocks.Selection(fn_ref, index=index)
    fn = building_blocks.Lambda(fn_ref.name, fn_ref.type_signature, sel)
    intrinsic = create_federated_map_or_apply(fn, value_ref)
    elements.append((name, intrinsic))
  result = building_blocks.Struct(
      elements,
      value.type_signature.member.python_container,  # pytype: disable=attribute-error
  )
  symbols = ((value_ref.name, value),)
  return building_blocks.Block(symbols, result)


def create_federated_value(
    value: building_blocks.ComputationBuildingBlock,
    placement: placements.PlacementLiteral,
) -> building_blocks.Call:
  r"""Creates a called federated value.

            Call
           /    \
  Intrinsic      Comp

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.
    placement: A `placements.PlacementLiteral` to use as the placement.

  Returns:
    A `building_blocks.Call`.
  """
  if placement is placements.CLIENTS:
    uri = intrinsic_defs.FEDERATED_VALUE_AT_CLIENTS.uri
  elif placement is placements.SERVER:
    uri = intrinsic_defs.FEDERATED_VALUE_AT_SERVER.uri
  else:
    raise TypeError('Unsupported placement {}.'.format(placement))
  result_type = computation_types.FederatedType(
      value.type_signature, placement, all_equal=True
  )
  intrinsic_type = computation_types.FunctionType(
      value.type_signature, result_type
  )
  intrinsic = building_blocks.Intrinsic(uri, intrinsic_type)
  return building_blocks.Call(intrinsic, value)


def _check_placements(placement_values: set[placements.PlacementLiteral]):
  """Checks if the placements of the values being zipped are compatible."""
  if not placement_values:
    raise TypeError(
        'federated_zip is only supported on nested structures '
        'containing at least one FederatedType, but none were '
        'found.'
    )
  elif len(placement_values) > 1:
    placement_list = ', '.join(placement.name for placement in placement_values)
    raise TypeError(
        'federated_zip requires all nested FederatedTypes to '
        'have the same placement, but values placed at '
        f'{placement_list} were found.'
    )


def create_federated_zip(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called federated zip.

  This function accepts a value whose type signature is a (potentially) nested
  tuple structure of federated values all with the same placement, and uses
  one of the federated_zip intrinsics (at client or at server) to promote the
  placement to the highest level. E.g., A value of type '<A@S, <<B@S>, C@S>>'
  would be mapped to a value of type '<A, <<B>, C>>@S'.

  Args:
    value: A `building_blocks.ComputationBuildingBlock` with a `type_signature`
      of type `computation_types.StructType` that may contain other nested
      `computation_types.StructTypes` bottoming out in at least one element of
      type `computation_Types.FederatedType`. These federated types must be at
      the same placement.

  Returns:
    A `building_blocks.Call` whose type signature is now a federated
      `computation_types.StructType`, placed at the same placement as the
      leaves of `value`.

  Raises:
    TypeError: If any of the types do not match.
    ValueError: If `value` does not contain any elements.
  """
  if not isinstance(value.type_signature, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {value.type_signature}.'
    )

  all_placements = set()

  def _record_placements(type_signature: computation_types.Type):
    """Records the placements in `type_signature` to `all_placements`."""
    if isinstance(type_signature, computation_types.FederatedType):
      all_placements.add(type_signature.placement)
    elif isinstance(type_signature, computation_types.StructType):
      for i, _ in enumerate(type_signature):
        _record_placements(type_signature[i])
    else:
      raise TypeError(
          'Expected type signatures consisting of structures of StructType '
          'bottoming out in FederatedType, found: \n{}'.format(type_signature)
      )

  _record_placements(value.type_signature)
  _check_placements(all_placements)
  placement = all_placements.pop()
  if placement is placements.CLIENTS:
    uri = intrinsic_defs.FEDERATED_ZIP_AT_CLIENTS.uri
  elif placement is placements.SERVER:
    uri = intrinsic_defs.FEDERATED_ZIP_AT_SERVER.uri
  else:
    raise TypeError('Unsupported placement {}.'.format(placement))

  def normalize_all_equals(element_type):
    if (
        isinstance(element_type, computation_types.FederatedType)
        and element_type.placement is placements.CLIENTS
        and element_type.all_equal
    ):
      return (
          computation_types.FederatedType(
              element_type.member, placements.CLIENTS
          ),
          True,
      )
    return element_type, False

  normalized_input_type, _ = type_transformations.transform_type_postorder(
      value.type_signature, normalize_all_equals
  )

  unplaced_output_type = type_transformations.strip_placement(
      value.type_signature
  )
  output_type = computation_types.FederatedType(unplaced_output_type, placement)
  intrinsic_type = computation_types.FunctionType(
      normalized_input_type, output_type
  )
  intrinsic = building_blocks.Intrinsic(uri, intrinsic_type)
  return building_blocks.Call(intrinsic, value)


def create_sequence_map(
    fn: building_blocks.ComputationBuildingBlock,
    arg: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called sequence map.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp]

  Args:
    fn: A `building_blocks.ComputationBuildingBlock` to use as the function.
    arg: A `building_blocks.ComputationBuildingBlock` to use as the argument.

  Returns:
    A `building_blocks.Call`.
  """
  result_type = computation_types.SequenceType(fn.type_signature.result)  # pytype: disable=attribute-error
  intrinsic_type = computation_types.FunctionType(
      (fn.type_signature, arg.type_signature), result_type
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.SEQUENCE_MAP.uri, intrinsic_type
  )
  values = building_blocks.Struct((fn, arg))
  return building_blocks.Call(intrinsic, values)


def create_sequence_reduce(
    value: building_blocks.ComputationBuildingBlock,
    zero: building_blocks.ComputationBuildingBlock,
    op: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called sequence reduce.

            Call
           /    \
  Intrinsic      Tuple
                 |
                 [Comp, Comp, Comp]

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.
    zero: A `building_blocks.ComputationBuildingBlock` to use as the initial
      value.
    op: A `building_blocks.ComputationBuildingBlock` to use as the op function.

  Returns:
    A `building_blocks.Call`.
  """
  op_parameter_type = computation_types.StructType([
      zero.type_signature,
      value.type_signature.element,  # pytype: disable=attribute-error
  ])
  op = _unname_fn_parameter(op, op_parameter_type)
  intrinsic_type = computation_types.FunctionType(
      (
          value.type_signature,
          zero.type_signature,
          op.type_signature,
      ),
      op.type_signature.result,  # pytype: disable=attribute-error
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.SEQUENCE_REDUCE.uri, intrinsic_type
  )
  values = building_blocks.Struct((value, zero, op))
  return building_blocks.Call(intrinsic, values)


def create_sequence_sum(
    value: building_blocks.ComputationBuildingBlock,
) -> building_blocks.Call:
  r"""Creates a called sequence sum.

            Call
           /    \
  Intrinsic      Comp

  Args:
    value: A `building_blocks.ComputationBuildingBlock` to use as the value.

  Returns:
    A `building_blocks.Call`.
  """
  intrinsic_type = computation_types.FunctionType(
      value.type_signature,
      value.type_signature.element,  # pytype: disable=attribute-error
  )
  intrinsic = building_blocks.Intrinsic(
      intrinsic_defs.SEQUENCE_SUM.uri, intrinsic_type
  )
  return building_blocks.Call(intrinsic, value)


def _create_naming_function(
    tuple_type_to_name: computation_types.StructType,
    names_to_add,
    container_type,
):
  """Private function to construct lambda naming a given tuple type.

  Args:
    tuple_type_to_name: Instance of `computation_types.StructType`, the type of
      the argument which we wish to name.
    names_to_add: Python `list` or `tuple`, the names we wish to give to
      `tuple_type_to_name`.
    container_type: Optional Python container type to associate with the
      resulting tuple.

  Returns:
    An instance of `building_blocks.Lambda` representing a function
    which will take an argument of type `tuple_type_to_name` and return a tuple
    with the same elements, but with names in `names_to_add` attached.

  Raises:
    ValueError: If `tuple_type_to_name` and `names_to_add` have different
    lengths.
  """
  if len(names_to_add) != len(tuple_type_to_name):  # pytype: disable=wrong-arg-types
    raise ValueError(
        'Number of elements in `names_to_add` must match number of element in '
        'the named tuple type `tuple_type_to_name`; here, `names_to_add` has '
        '{} elements and `tuple_type_to_name` has {}.'.format(
            len(names_to_add),  # pytype: disable=wrong-arg-types
            len(tuple_type_to_name),  # pytype: disable=wrong-arg-types
        )
    )
  naming_lambda_arg = building_blocks.Reference('x', tuple_type_to_name)

  def _create_struct_element(i):
    return (
        names_to_add[i],
        building_blocks.Selection(naming_lambda_arg, index=i),
    )

  named_result = building_blocks.Struct(
      [_create_struct_element(k) for k in range(len(names_to_add))],
      container_type,
  )
  return building_blocks.Lambda(
      'x', naming_lambda_arg.type_signature, named_result
  )


def create_named_tuple(
    comp: building_blocks.ComputationBuildingBlock,
    names: Sequence[Optional[str]],
    container_type=None,
) -> building_blocks.ComputationBuildingBlock:
  """Creates a computation that applies `names` to `comp`.

  Args:
    comp: A `building_blocks.ComputationBuildingBlock` with a `type_signature`
      of type `computation_types.StructType`.
    names: Python `tuple` or `list` containing instances of type `str` or
      `None`, the names to apply to `comp`.
    container_type: Optional Python container type to associated with the
      resulting tuple.

  Returns:
    A `building_blocks.ComputationBuildingBlock` representing a
    tuple with the elements from `comp` and the names from `names` attached to
    the `type_signature` of those elements.
  """
  if not isinstance(comp.type_signature, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {comp.type_signature}.'
    )

  fn = _create_naming_function(comp.type_signature, names, container_type)
  return building_blocks.Call(fn, comp)


def zip_to_match_type(
    *,
    comp_to_zip: building_blocks.ComputationBuildingBlock,
    target_type: computation_types.Type,
) -> Optional[building_blocks.ComputationBuildingBlock]:
  """Zips computation argument to match target type.

  This function will apply the appropriate federated zips to match `comp_to_zip`
  to the requested type `target_type`, subject to a few caveats. We will
  traverse `computation_types.StructTypes` to match types, so for example we
  would zip `<<T@P, R@P>>` to match `<<T, R>@P>`, but we will not traverse
  `computation_types.FunctionTypes`. Therefore we would not apply a zip to the
  parameter of `(<<T@P, R@P>> -> Q)` to match (<<T, R>@P> -> Q).

  If zipping in this manner cannot match the type of `comp_to_zip` to
  `target_type`, `None` will be returned.

  Args:
    comp_to_zip: Instance of `building_blocks.ComputationBuildingBlock` to
      traverse and attempt to zip to match `target_type`.
    target_type: The type to target when traversing and zipping `comp_to_zip`.

  Returns:
    Either a potentially transformed version of `comp_to_zip` or `None`,
    depending on whether inserting a zip according to the semantics above
    can transformed `comp_to_zip` to the requested type.
  """

  def _can_be_zipped_into(
      source_type: computation_types.Type, target_type: computation_types.Type
  ) -> bool:
    """Indicates possibility of the transformation `zip_to_match_type`."""

    def _struct_can_be_zipped_to_federated(
        struct_type: computation_types.StructType,
        federated_type: computation_types.FederatedType,
    ) -> bool:
      placements_encountered = set()

      def _remove_placement(
          subtype: computation_types.Type,
      ) -> tuple[computation_types.Type, bool]:
        if isinstance(subtype, computation_types.FederatedType):
          placements_encountered.add(subtype.placement)
          return subtype.member, True
        return subtype, False

      unplaced_struct, _ = type_transformations.transform_type_postorder(
          struct_type, _remove_placement
      )
      if not (
          all(x is federated_type.placement for x in placements_encountered)
      ):
        return False
      if (
          federated_type.placement is placements.CLIENTS
          and federated_type.all_equal
      ):
        # There is no all-equal clients zip; return false.
        return False
      return federated_type.member.is_assignable_from(unplaced_struct)

    def _struct_elem_zippable(
        source_name, source_element, target_name, target_element
    ):
      return _can_be_zipped_into(
          source_element, target_element
      ) and source_name in (target_name, None)

    if isinstance(source_type, computation_types.StructType):
      if isinstance(target_type, computation_types.FederatedType):
        return _struct_can_be_zipped_to_federated(source_type, target_type)
      elif isinstance(target_type, computation_types.StructType):
        elements_zippable = []
        for (s_name, s_el), (t_name, t_el) in zip(
            source_type.items(),
            target_type.items(),
        ):
          elements_zippable.append(
              _struct_elem_zippable(s_name, s_el, t_name, t_el)
          )
        return all(elements_zippable)
    else:
      return target_type.is_assignable_from(source_type)

  def _zip_to_match(
      *,
      source: building_blocks.ComputationBuildingBlock,
      target_type: computation_types.Type,
  ):
    if isinstance(target_type, computation_types.FederatedType) and isinstance(
        source.type_signature, computation_types.StructType
    ):
      return create_federated_zip(source)
    elif isinstance(target_type, computation_types.StructType) and isinstance(
        source.type_signature, computation_types.StructType
    ):
      zipped_elements = []
      # Bind a reference to the source to prevent duplication in the AST.
      ref_name = next(unique_name_generator(source))
      ref_to_source = building_blocks.Reference(ref_name, source.type_signature)
      for idx, ((_, t_el), (s_name, _)) in enumerate(
          zip(
              target_type.items(),
              source.type_signature.items(),
          )
      ):
        s_selection = building_blocks.Selection(ref_to_source, index=idx)
        zipped_elements.append(
            (s_name, _zip_to_match(source=s_selection, target_type=t_el))
        )
        # Insert binding above the constructed structure.
        return building_blocks.Block(
            [(ref_name, source)], building_blocks.Struct(zipped_elements)
        )
    else:
      # No zipping to be done here.
      return source

  if target_type.is_assignable_from(comp_to_zip.type_signature):
    # No zipping needs to be done; return directly.
    return comp_to_zip
  elif _can_be_zipped_into(comp_to_zip.type_signature, target_type):
    return _zip_to_match(source=comp_to_zip, target_type=target_type)
  else:
    # Zipping cannot be performed here.
    return None



## Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Concurrency utilities for use with Python `async`."""

import asyncio
import threading

from federated_language.common_libs import tracing


class AsyncThreadRunner:
  """Class which bridges async and synchronous synchronous interfaces.

  This class serves as a resource and logic container, starting an event loop
  in a separate thread and managing dispatching of coroutine functions to this
  event loop in both synchronous and asynchronous interfaces.

  There are two main uses of this class. First, this class can be used to wrap
  interfaces which use `asyncio` in a synchronous 'run this coroutine'
  interface in a manner which is compatible with integrating with other async
  libraries. This feature is generally useful for backwards-compatibility (e.g.,
  introducing asyncio in some component which sits on top of the synchronous
  function calls this interface exposes), but should generally be viewed as
  suboptimal--it is preferable in a situation like this to simply expose the
  underlying async interfaces.

  Second, this class can be used to delegate asynchronous work from one thread
  to another, using its asynchronous interface.
  """

  def __init__(self):
    self._event_loop = asyncio.new_event_loop()
    self._event_loop.set_task_factory(
        tracing.propagate_trace_context_task_factory
    )

    def target_fn():
      self._event_loop.run_forever()

    self._thread = threading.Thread(target=target_fn, daemon=True)
    self._thread.start()

    def finalizer(loop, thread):
      loop.call_soon_threadsafe(loop.stop)
      thread.join()

    self._finalizer = finalizer

  def __del__(self):
    self._finalizer(self._event_loop, self._thread)

  def run_coro_and_return_result(self, coro):
    """Runs coroutine in the managed event loop, returning the result."""
    future = asyncio.run_coroutine_threadsafe(coro, self._event_loop)
    return future.result()

  async def await_coro_and_return_result(self, coro):
    """Runs coroutine in the managed event loop, returning the result."""
    return await asyncio.wrap_future(
        asyncio.run_coroutine_threadsafe(coro, self._event_loop)
    )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions of all intrinsic for use within the system."""

import enum
from typing import Optional

from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_factory
import numpy as np

_intrinsic_registry = {}


@enum.unique
class BroadcastKind(enum.Enum):
  DEFAULT = 1
  SECURE = 2


@enum.unique
class AggregationKind(enum.Enum):
  DEFAULT = 1
  SECURE = 2


class IntrinsicDef:
  """Represents the definition of an intrinsic.

  This class represents the ultimate source of ground truth about what kinds of
  intrinsics exist and what their types are. To be consuled by all other code
  that deals with intrinsics.
  """

  def __init__(
      self,
      name: str,
      uri: str,
      type_signature: computation_types.Type,
      aggregation_kind: Optional[AggregationKind] = None,
      broadcast_kind: Optional[BroadcastKind] = None,
  ):
    """Constructs a definition of an intrinsic.

    Args:
      name: The short human-friendly name of this intrinsic.
      uri: The URI of this intrinsic.
      type_signature: The type of the intrinsic.
      aggregation_kind: Optional kind of aggregation performed by calls.
      broadcast_kind: Optional kind of broadcast performed by calls.
    """
    self._name = str(name)
    self._uri = str(uri)
    self._type_signature = type_signature
    self._aggregation_kind = aggregation_kind
    self._broadcast_kind = broadcast_kind
    _intrinsic_registry[str(uri)] = self

  @property
  def name(self):
    return self._name

  @property
  def uri(self):
    return self._uri

  @property
  def type_signature(self):
    return self._type_signature

  @property
  def aggregation_kind(self) -> Optional[AggregationKind]:
    return self._aggregation_kind

  @property
  def broadcast_kind(self) -> Optional[BroadcastKind]:
    return self._broadcast_kind

  def __str__(self):
    return self._name

  def __repr__(self):
    return "IntrinsicDef('{}')".format(self._uri)


# Computes an aggregate of client items (the first, {T}@CLIENTS-typed parameter)
# using a multi-stage process in which client items are first partially
# aggregated at an intermediate layer, then the partial aggregates are further
# combined, and finally projected into the result. This multi-stage process is
# parameterized by a four-part aggregation interface that consists of the
# following:
# a) The 'zero' in the algebra used at the initial stage (partial aggregation),
#    This is the second, U-typed parameter.
# b) The operator that accumulates T-typed client items into the U-typed partial
#    aggregates. This is the third, (<U,T>->U)-typed parameter.
# c) The operator that combines pairs of U-typed partial aggregates. This is the
#    fourth, (<U,U>->U)-typed parameter.
# d) The operator that projects the top-level aggregate into the final result.
#    This is the fifth, (U->R)-typed parameter.
#
# Conceptually, given a new literal INTERMEDIATE_AGGREGATORS in a single-layer
# aggregation architecture, one could define this intrinsic in terms of generic
# intrinsics defined above, as follows.
#
# @federated_computation
# def federated_aggregate(x, zero, accumulate, merge, report):
#   a = generic_partial_reduce(x, zero, accumulate, INTERMEDIATE_AGGREGATORS)
#   b = generic_reduce(a, zero, merge, SERVER)
#   c = generic_map(report, b)
#   return c
#
# Actual implementations might vary.
#
# Type signature: <{T}@CLIENTS,U,(<U,T>->U),(<U,U>->U),(U->R)> -> R@SERVER
FEDERATED_AGGREGATE = IntrinsicDef(
    'FEDERATED_AGGREGATE',
    'federated_aggregate',
    computation_types.FunctionType(
        parameter=[
            computation_types.FederatedType(
                computation_types.AbstractType('T'), placements.CLIENTS
            ),
            computation_types.AbstractType('U'),
            type_factory.reduction_op(
                computation_types.AbstractType('U'),
                computation_types.AbstractType('T'),
            ),
            type_factory.binary_op(computation_types.AbstractType('U')),
            computation_types.FunctionType(
                computation_types.AbstractType('U'),
                computation_types.AbstractType('R'),
            ),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('R'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.DEFAULT,
)

# Applies a given function to a value on the server.
#
# Type signature: <(T->U),T@SERVER> -> U@SERVER
FEDERATED_APPLY = IntrinsicDef(
    'FEDERATED_APPLY',
    'federated_apply',
    computation_types.FunctionType(
        parameter=[
            computation_types.FunctionType(
                computation_types.AbstractType('T'),
                computation_types.AbstractType('U'),
            ),
            computation_types.FederatedType(
                computation_types.AbstractType('T'), placements.SERVER
            ),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('U'), placements.SERVER
        ),
    ),
)

# Broadcasts a server item to all clients.
#
# Type signature: T@SERVER -> T@CLIENTS
FEDERATED_BROADCAST = IntrinsicDef(
    'FEDERATED_BROADCAST',
    'federated_broadcast',
    computation_types.FunctionType(
        parameter=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'),
            placements.CLIENTS,
            all_equal=True,
        ),
    ),
    broadcast_kind=BroadcastKind.DEFAULT,
)

# Evaluates a function at the clients.
#
# Type signature: (() -> T) -> {T}@CLIENTS
FEDERATED_EVAL_AT_CLIENTS = IntrinsicDef(
    'FEDERATED_EVAL_AT_CLIENTS',
    'federated_eval_at_clients',
    computation_types.FunctionType(
        parameter=computation_types.FunctionType(
            None, computation_types.AbstractType('T')
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.CLIENTS
        ),
    ),
)

# Evaluates a function at the server.
#
# Type signature: (() -> T) -> T@SERVER
FEDERATED_EVAL_AT_SERVER = IntrinsicDef(
    'FEDERATED_EVAL_AT_SERVER',
    'federated_eval_at_server',
    computation_types.FunctionType(
        parameter=computation_types.FunctionType(
            None, computation_types.AbstractType('T')
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
)

# Maps member constituents of a client value pointwise using a given mapping
# function that operates independently on each client.
#
# Type signature: <(T->U),{T}@CLIENTS> -> {U}@CLIENTS
FEDERATED_MAP = IntrinsicDef(
    'FEDERATED_MAP',
    'federated_map',
    computation_types.FunctionType(
        parameter=[
            computation_types.FunctionType(
                computation_types.AbstractType('T'),
                computation_types.AbstractType('U'),
            ),
            computation_types.FederatedType(
                computation_types.AbstractType('T'), placements.CLIENTS
            ),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('U'), placements.CLIENTS
        ),
    ),
)

# Maps member constituents of a client all equal value pointwise using a given
# mapping function that operates independently on each client, as a result of
# this independence, the value is only garunteed to be all equal if the function
# is deterministic.
#
# Type signature: <(T->U),T@CLIENTS> -> U@CLIENTS
FEDERATED_MAP_ALL_EQUAL = IntrinsicDef(
    'FEDERATED_MAP_ALL_EQUAL',
    'federated_map_all_equal',
    computation_types.FunctionType(
        parameter=[
            computation_types.FunctionType(
                computation_types.AbstractType('T'),
                computation_types.AbstractType('U'),
            ),
            computation_types.FederatedType(
                computation_types.AbstractType('T'),
                placements.CLIENTS,
                all_equal=True,
            ),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('U'),
            placements.CLIENTS,
            all_equal=True,
        ),
    ),
)

# Computes a simple (equally weighted) mean of client items. Only supported
# for numeric tensor types, or composite structures made up of numeric types.
#
# Type signature: {T}@CLIENTS -> T@SERVER
FEDERATED_MEAN = IntrinsicDef(
    'FEDERATED_MEAN',
    'federated_mean',
    computation_types.FunctionType(
        parameter=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.CLIENTS
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.DEFAULT,
)

# Computes the min of client values on the server. Only supported for numeric
# types, or nested structures made up of numeric computation_types.
#
# Type signature: {T}@CLIENTS -> T@SERVER
FEDERATED_MIN = IntrinsicDef(
    'FEDERATED_MIN',
    'federated_min',
    computation_types.FunctionType(
        parameter=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.CLIENTS
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.DEFAULT,
)

# Computes the max of client values on the server. Only supported for numeric
# types, or nested structures made up of numeric computation_types.
#
# Type signature: {T}@CLIENTS -> T@SERVER
FEDERATED_MAX = IntrinsicDef(
    'FEDERATED_MAX',
    'federated_max',
    computation_types.FunctionType(
        parameter=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.CLIENTS
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.DEFAULT,
)

# Computes the sum of client values on the server, securely. Only supported for
# integers or nested structures of integers.
#
# Type signature: <{V}@CLIENTS,M> -> V@SERVER
FEDERATED_SECURE_SUM = IntrinsicDef(
    'FEDERATED_SECURE_SUM',
    'federated_secure_sum',
    computation_types.FunctionType(
        parameter=[
            computation_types.FederatedType(
                computation_types.AbstractType('V'), placements.CLIENTS
            ),
            computation_types.AbstractType('M'),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('V'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.SECURE,
)

# Computes the sum of client values on the server, securely. Only supported for
# integers or nested structures of integers.
#
# Type signature: <{V}@CLIENTS,B> -> V@SERVER
FEDERATED_SECURE_SUM_BITWIDTH = IntrinsicDef(
    'FEDERATED_SECURE_SUM_BITWIDTH',
    'federated_secure_sum_bitwidth',
    computation_types.FunctionType(
        parameter=[
            computation_types.FederatedType(
                computation_types.AbstractType('V'), placements.CLIENTS
            ),
            computation_types.AbstractType('B'),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('V'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.SECURE,
)

_SELECT_TYPE = computation_types.FunctionType(
    parameter=[
        computation_types.FederatedType(
            computation_types.AbstractType('Ks'), placements.CLIENTS
        ),  # client_keys
        computation_types.FederatedType(np.int32, placements.SERVER),  # max_key
        computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),  # server_state
        computation_types.FunctionType(
            [computation_types.AbstractType('T'), np.int32],
            computation_types.AbstractType('U'),
        ),  # select_fn
    ],
    result=computation_types.FederatedType(
        computation_types.SequenceType(computation_types.AbstractType('U')),
        placements.CLIENTS,
    ),
)

# Distributes server values to clients based on client keys.
FEDERATED_SELECT = IntrinsicDef(
    'FEDERATED_SELECT',
    'federated_select',
    _SELECT_TYPE,
    broadcast_kind=BroadcastKind.DEFAULT,
)

# Securely distributes server values to clients based on private client keys.
FEDERATED_SECURE_SELECT = IntrinsicDef(
    'FEDERATED_SECURE_SELECT',
    'federated_secure_select',
    _SELECT_TYPE,
    broadcast_kind=BroadcastKind.SECURE,
)

# Computes the sum of client values on the server. Only supported for numeric
# types, or nested structures made up of numeric computation_types.
#
# Type signature: {T}@CLIENTS -> T@SERVER
FEDERATED_SUM = IntrinsicDef(
    'FEDERATED_SUM',
    'federated_sum',
    computation_types.FunctionType(
        parameter=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.CLIENTS
        ),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.DEFAULT,
)

# Places a value at the clients.
#
# Type signature: T -> T@CLIENTS
FEDERATED_VALUE_AT_CLIENTS = IntrinsicDef(
    'FEDERATED_VALUE_AT_CLIENTS',
    'federated_value_at_clients',
    computation_types.FunctionType(
        parameter=computation_types.AbstractType('T'),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.CLIENTS, True
        ),
    ),
)

# Places a value at the server.
#
# Type signature: T -> T@SERVER
FEDERATED_VALUE_AT_SERVER = IntrinsicDef(
    'FEDERATED_VALUE_AT_SERVER',
    'federated_value_at_server',
    computation_types.FunctionType(
        parameter=computation_types.AbstractType('T'),
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
)

# Computes a weighted mean of client items. Only supported for numeric tensor
# types, or composite structures made up of numeric types. Weights must be
# simple scalar numeric (integer or floating point, not complex) tensor types.
# The types of weights and values must be compatible, i.e., multiplying and
# dividing member constituents of the value by weights should yield results of
# the same type as the type of these member constituents being weighted. Thus,
# for example, one may not supply values containing `np.int32`` tensors, as the
# result of weighting such values is of a floating-point type. Casting must be
# injected, where appropriate, by the compiler before invoking this intrinsic.
#
# Type signature: <{T}@CLIENTS,{U}@CLIENTS> -> T@SERVER
FEDERATED_WEIGHTED_MEAN = IntrinsicDef(
    'FEDERATED_WEIGHTED_MEAN',
    'federated_weighted_mean',
    computation_types.FunctionType(
        parameter=[
            computation_types.FederatedType(
                computation_types.AbstractType('T'), placements.CLIENTS
            ),
            computation_types.FederatedType(
                computation_types.AbstractType('U'), placements.CLIENTS
            ),
        ],
        result=computation_types.FederatedType(
            computation_types.AbstractType('T'), placements.SERVER
        ),
    ),
    aggregation_kind=AggregationKind.DEFAULT,
)

# Zips a tuple of two federated types into a federated tuple.
#
# Type signature: T -> U@CLIENTS
# where `T` is a structure of client-placed values.
# `U` must be identical to `T` with federated placement removed.
FEDERATED_ZIP_AT_CLIENTS = IntrinsicDef(
    'FEDERATED_ZIP_AT_CLIENTS',
    'federated_zip_at_clients',
    computation_types.FunctionType(
        parameter=computation_types.AbstractType('T'),
        result=computation_types.FederatedType(
            computation_types.AbstractType('U'), placements.CLIENTS
        ),
    ),
)
# Type signature: T -> U@SERVER
# where `T` is a structure of server-placed values.
# `U` must be identical to `T` with federated placement removed.
FEDERATED_ZIP_AT_SERVER = IntrinsicDef(
    'FEDERATED_ZIP_AT_SERVER',
    'federated_zip_at_server',
    computation_types.FunctionType(
        parameter=computation_types.AbstractType('T'),
        result=computation_types.FederatedType(
            computation_types.AbstractType('U'), placements.SERVER
        ),
    ),
)

# Generic plus operator that accepts a variety of types in federated computation
# context. The range of types 'T' supported to be defined. It should work in a
# natural manner for tensors, tuples, federated types, possibly sequences, and
# maybe even functions (although it's unclear if such generality is desirable).
#
# Type signature: <T,T> -> T
GENERIC_PLUS = IntrinsicDef(
    'GENERIC_PLUS',
    'generic_plus',
    type_factory.binary_op(computation_types.AbstractType('T')),
)

# Performs pointwise TensorFlow division on its arguments.
# The type signature of generic divide is determined by TensorFlow's set of
# implicit type equations. For example, dividing `int32` by `int32` in TF
# generates a tensor of type `float64`. There is therefore more structure than
# is suggested by the type signature `<T,T> -> U`.
#
# Type signature: <T,T> -> U
GENERIC_DIVIDE = IntrinsicDef(
    'GENERIC_DIVIDE',
    'generic_divide',
    computation_types.FunctionType(
        [
            computation_types.AbstractType('T'),
            computation_types.AbstractType('T'),
        ],
        computation_types.AbstractType('U'),
    ),
)

# Performs pointwise TensorFlow multiplication on its arguments.
#
# Type signature: <T,T> -> T
GENERIC_MULTIPLY = IntrinsicDef(
    'GENERIC_MULTIPLY',
    'generic_multiply',
    computation_types.FunctionType(
        [computation_types.AbstractType('T')] * 2,
        computation_types.AbstractType('T'),
    ),
)
# Generic zero operator that represents zero-filled values of diverse types (to
# be defined, but generally similar to that supported by GENERIC_ADD).
#
# Type signature: T
GENERIC_ZERO = IntrinsicDef(
    'GENERIC_ZERO', 'generic_zero', computation_types.AbstractType('T')
)

# Maps elements of a sequence using a given mapping function that operates
# independently on each element.
#
# Type signature: <(T->U),T*> -> U*
SEQUENCE_MAP = IntrinsicDef(
    'SEQUENCE_MAP',
    'sequence_map',
    computation_types.FunctionType(
        parameter=[
            computation_types.FunctionType(
                computation_types.AbstractType('T'),
                computation_types.AbstractType('U'),
            ),
            computation_types.SequenceType(computation_types.AbstractType('T')),
        ],
        result=computation_types.SequenceType(
            computation_types.AbstractType('U')
        ),
    ),
)

# Reduces a sequence using a given 'zero' in the algebra (i.e., the result of
# reducing an empty sequence) and a given reduction operator with the signature
# U,T->U that incorporates a single T-typed element into a U-typed result of
# partial reduction. In the special case of T = U, this corresponds to the
# classical notion of reduction of a set using a commutative associative binary
# operator. The generalized reduction operator (with T != U) must yield the same
# results when repeatedly applied on sequences of elements in any order.
#
# Type signature: <T*,U,(<U,T>->U)> -> U
SEQUENCE_REDUCE = IntrinsicDef(
    'SEQUENCE_REDUCE',
    'sequence_reduce',
    computation_types.FunctionType(
        parameter=[
            computation_types.SequenceType(computation_types.AbstractType('T')),
            computation_types.AbstractType('U'),
            type_factory.reduction_op(
                computation_types.AbstractType('U'),
                computation_types.AbstractType('T'),
            ),
        ],
        result=computation_types.AbstractType('U'),
    ),
)

# Computes the sum of values in a sequence. Only supported for numeric types
# or nested structures made up of numeric types.
#
# Type signature: T* -> T
SEQUENCE_SUM = IntrinsicDef(
    'SEQUENCE_SUM',
    'sequence_sum',
    computation_types.FunctionType(
        parameter=computation_types.SequenceType(
            computation_types.AbstractType('T')
        ),
        result=computation_types.AbstractType('T'),
    ),
)


def uri_to_intrinsic_def(uri) -> Optional[IntrinsicDef]:
  return _intrinsic_registry.get(uri)


# TODO: b/254770431 - Add documentation explaining the implications of setting
# broadcast_kind for an intrinsic.
def get_broadcast_intrinsics() -> list[IntrinsicDef]:
  return [
      intrinsic
      for intrinsic in _intrinsic_registry.values()
      if intrinsic.broadcast_kind
  ]


# TODO: b/254770431 - Add documentation explaining the implications of setting
# aggregation_kind for an intrinsic.
def get_aggregation_intrinsics() -> list[IntrinsicDef]:
  return [
      intrinsic
      for intrinsic in _intrinsic_registry.values()
      if intrinsic.aggregation_kind
  ]



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The implementation of a context to use in building federated computations."""

from typing import Optional

from federated_language.compiler import building_blocks
from federated_language.context_stack import context_stack_impl
from federated_language.context_stack import symbol_binding_context
from federated_language.federated_context import value_impl
from federated_language.types import computation_types
from federated_language.types import type_conversions


class FederatedComputationContext(
    symbol_binding_context.SymbolBindingContext[
        building_blocks.ComputationBuildingBlock,
        building_blocks.Reference,
    ]
):
  """The context for building federated computations.

  This context additionally holds a list of symbols which are bound to
  `building_block.ComputationBuildingBlocks` during construction of
  `federated_language.Values`, and which respect identical semantics to the
  binding of locals
  in `building_blocks.Blocks`.

  Any `federated_language.Value` constructed in this context may add such a
  symbol binding,
  and thereafter refer to the returned reference in place of the bound
  computation. It is then the responsibility of the installer of this context
  to ensure that the symbols bound during the `federated_language.Value`
  construction process
  are appropriately packaged in the result.
  """

  def __init__(
      self,
      context_stack: context_stack_impl.ContextStack,
      suggested_name: Optional[str] = None,
      parent: Optional['FederatedComputationContext'] = None,
  ):
    """Creates this context.

    Args:
      context_stack: The context stack to use.
      suggested_name: The optional suggested name of the context, a string. It
        may be modified to make it different from the names of any of the
        ancestors on the context stack.
      parent: The optional parent context. If not `None`, it must be an instance
        of `FederatedComputationContext`.
    """
    if suggested_name is None or not suggested_name:
      suggested_name = 'FEDERATED'
    ancestor = parent
    ancestor_names = set()
    while ancestor is not None:
      ancestor_names.add(ancestor.name)
      ancestor = ancestor.parent
    name = suggested_name
    name_count = 0
    while name in ancestor_names:
      name_count = name_count + 1
      name = '{}_{}'.format(suggested_name, name_count)
    self._context_stack = context_stack
    self._parent = parent
    self._name = name
    self._symbol_bindings = []
    self._next_symbol_val = 0

  @property
  def name(self):
    return self._name

  @property
  def parent(self):
    return self._parent

  def bind_computation_to_reference(
      self, comp: building_blocks.ComputationBuildingBlock
  ) -> building_blocks.Reference:
    """Binds a computation to a symbol, returns a reference to this binding."""
    name = 'fc_{name}_symbol_{val}'.format(
        name=self._name, val=self._next_symbol_val
    )
    self._next_symbol_val += 1
    self._symbol_bindings.append((name, comp))
    ref = building_blocks.Reference(name, comp.type_signature)
    return ref

  @property
  def symbol_bindings(
      self,
  ) -> list[tuple[str, building_blocks.ComputationBuildingBlock]]:
    return self._symbol_bindings

  def invoke(self, comp, arg):
    fn = value_impl.to_value(comp, type_spec=None)
    type_spec = fn.type_signature
    if not isinstance(type_spec, computation_types.FunctionType):
      raise ValueError(
          f'Expected {type_spec} to be a `federated_language.FunctionType`.'
      )
    if arg is not None:
      if type_spec.parameter is None:
        raise ValueError(f'Expected no arguments, found {arg}.')
      arg = value_impl.to_value(
          arg, type_spec=type_spec.parameter, zip_if_needed=True
      )
      result = fn(arg)
    else:
      if type_spec.parameter is not None:
        raise ValueError(
            f'Expected an argument of type {type_spec.parameter}, found none.'
        )
      result = fn()

    value_type = type_conversions.infer_type(result)
    if not type_spec.result.is_assignable_from(value_type):
      raise computation_types.TypeNotAssignableError(value_type, type_spec)

    return result
